from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import copy
from operator import attrgetter
import os
import pickle
import random
import time

import deap
import deap.tools
from deap.tools import HallOfFame
import deap.algorithms
import numpy as np
import six
from sklearn.base import BaseEstimator
from sklearn.feature_selection import SelectorMixin

import casm.learn


def initNRandomOn(container, n_features, n_features_init):
    """
  Initialize a random container of bool.

  Arguments
  ---------

    container: Type
      The type of container to return. Expects container(List) to be valid.

    n_features: int
      The total length of the container

    n_features_init: int
      The number of True elements in the container

  Returns
  -------

    result: container
      A container of the specified type with 'n_features' True and the rest False.
  """
    bits = [False] * n_features
    for i in random.sample(range(n_features), n_features_init):
        bits[i] = True
    return container(bits)


class Constraints(object):
    """
  Holds constraints on individuals in evolutionary algorithms.

  Attributes
  ----------

    n_features_min: int
      The minimum allowed number of selected features. Must be >=1.

    n_features_max: int or str
      The maximum allowed number of selected features. String "all" for no limit.

    fix_on: 1d array-like of int
      The indices of features to fix on

    fix_off: 1d array-like of int
      The indices of features to fix off

  """
    def __init__(self,
                 n_features_min=1,
                 n_features_max="all",
                 fix_on=np.array([], dtype=int),
                 fix_off=np.array([], dtype=int)):
        """
    Arguments
    ---------

      n_features_min: int, optional, default=1
        The minimum allowed number of selected features. Must be >=1.

      n_features_max: int or str, optional, default "all".
        The maximum allowed number of selected features. String "all" for no limit.

      fix_on: 1d array-like of int, optional, default=np.array([], dtype=int)
        The indices of features to fix on

      fix_off: 1d array-like of int, optional, default=np.array([], dtype=int)
        The indices of features to fix off

    """
        if n_features_min < 1:
            raise ValueError("n_features_min must be >= 1")
        if n_features_max != "all":
            n_features_max = int(n_features_max)
        self.n_features_min = n_features_min
        self.n_features_max = n_features_max
        self.fix_on = np.array(fix_on, dtype=int)
        self.fix_off = np.array(fix_off, dtype=int)

    def check(self, indiv):
        # first ensure fix_on and fix_off
        for i in self.fix_on:
            if not indiv[i]:
                return False

        for i in self.fix_off:
            if indiv[i]:
                return False

        Non = sum(indiv)
        if Non < self.n_features_min:
            return False
        if Non > self.n_features_max:
            return False

        return True


def restrict_constraints(constraints):
    """
  Returns a decorator that removes offspring that don't obey constraints.

  The decorator does:
    offspring = func(*args, **kargs)
    return [child for child in offspring if constraints.check(child)]

  Arguments
  ---------
    constraints: casm.learn.evolve.Constraints
      A Constraints object containing the requested constraints


  Returns
  -------

    decorator: function
      A decorator that can be used to wrap evolutionary algorithm methods that
      return a population to ensure that constraints remain satisfied.
  """
    def decorator(func):
        def wrapper(*args, **kargs):
            offspring = func(*args, **kargs)
            return [child for child in offspring if constraints.check(child)]

        return wrapper

    return decorator


def enforce_constraints(constraints):
    """
  Returns a decorator that ensures that individuals obey constraints.

  The decorator will first select on/off any features that are constrained by
  'fix_on' or 'fix_off'. Then, if the number of features selected is less than
  'n_features_min', features will be randomly turned on. Finally, if the number
  of features selected is less than 'n_features_max', features will be randomly
  turned off.

  Arguments
  ---------
    constraints: casm.learn.evolve.Constraints
      A Constraints object containing the requested constraints


  Returns
  -------

    decorator: function
      A decorator that can be used to wrap evolutionary algorithm methods that
      return a population to ensure that constraints remain satisfied.
  """
    def decorator(func):
        def wrapper(*args, **kargs):
            offspring = func(*args, **kargs)
            for child in offspring:

                # first ensure fix_on and fix_off
                for i in constraints.fix_on:
                    child[i] = True
                for i in constraints.fix_off:
                    child[i] = False

                # now make sure enough are 'on'
                Non = sum(child)
                if Non < constraints.n_features_min:
                    # get all 'off'
                    off = np.where(np.asarray(child) == False)[0]

                    # remove those that are fix_off
                    candidate = set(off).difference(set(constraints.fix_off))

                    # select enough of the candidates to reach the minimum
                    turn_on = random.sample(candidate,
                                            constraints.n_features_min - Non)
                    for i in turn_on:
                        child[i] = True

                # now make sure enough are 'off'
                if constraints.n_features_max != "all":
                    if Non > constraints.n_features_max:
                        # get all 'on'
                        on = np.where(np.asarray(child) == True)[0]

                        # remove those that are fix_on
                        candidate = set(on).difference(set(constraints.fix_on))

                        # select enough of the candidates to reach the maximum
                        turn_off = random.sample(
                            candidate, Non - constraints.n_features_max)
                        for i in turn_off:
                            child[i] = False

            return offspring

        return wrapper

    return decorator


def initialize_population(n_population, toolbox, filename=None, verbose=True):
    """
  if filename is not None and os.path.exists(filename):
    load initial population
    # may be List[Individual] or HallOfFame, which is converted to List[Individual]
  else:
    create random initial population of size n_population via toolbox.population
  """
    # load or construct initiali population

    if filename is not None and os.path.exists(filename):
        if verbose:
            print("Loading initial population:", filename)
        with open(filename, 'rb') as f:
            pop = pickle.load(f)
        if isinstance(pop, HallOfFame):
            # convert to List
            pop = [indiv for indiv in pop]
    else:
        if verbose:
            print("Constructing initial population")
        pop = toolbox.population(n_population)

    return pop


def save_population(pop, filename, verbose=False):
    if verbose:
        print("\nPickling population to:", filename)
    with open(filename, 'wb') as f:
        pickle.dump(pop, f, protocol=2)


def initialize_halloffame(filename=None, n_halloffame=25, verbose=False):
    hall = casm.learn.create_halloffame(n_halloffame)
    if verbose:
        print("# Hall of Fame size:", n_halloffame, "\n")

    if os.path.exists(filename):
        with open(filename, 'rb') as f:
            existing_hall = pickle.load(f)
        if verbose:
            print("Loading Hall of Fame:", filename)
        hall.update(existing_hall)
    return hall


def save_halloffame(hall, filename, verbose=False):
    if verbose:
        print("\nPickling Hall of Fame to:", filename)
    with open(filename, 'wb') as f:
        pickle.dump(hall, f, protocol=2)


def evaluate_all(pop, toolbox):
    """
  Evaluate and set fitness of all individual's with invalid fitness.

  Arguments
  ---------

    pop: iterable of individual
      Population to be evaluated

    toolbox: deap.base.Toolbox
      Contains methods used by deap during evolution. Expected to contain:

        toolbox.evaluate(indiv): To evaluate an individual's fitness

        toolbox.map(func, List[individual]): To map function executions

  Returns
  -------

    nevals: int
      Number of evaluations performed

  """
    # evaluate initial fitness
    invalid_ind = [ind for ind in pop if not ind.fitness.valid]
    fitnesses = toolbox.map(toolbox.evaluate, invalid_ind)
    for ind, fit in zip(invalid_ind, fitnesses):
        ind.fitness.values = fit
    return len(invalid_ind)


class EvolutionaryParams(object):
    """
  Holds parameters used by evolutionary algorithms.

  Attributes
  ----------

    n_population: int
      Population size.

    n_generation: int
      Number of generations to for each repitition. Results are saved between
      repetitions.

    n_repetition: int
      Number of repititions to perform.

    n_features_init: int
      Number of randomly selected features to initialize each individual with.

    pop_begin_filename: string
      Filename where the initial population is read from, if it exists.

    pop_end_filename: string
      Filename where the final population is saved.

    halloffame_filename: string
      Filename where a hall of fame is saved.

    n_halloffame: int
      Number of individuals to save in the hall of fame
  """
    def __init__(self,
                 n_population=100,
                 n_generation=10,
                 n_repetition=100,
                 n_features_init=1,
                 pop_begin_filename="population_begin.pkl",
                 pop_end_filename="population_end.pkl",
                 halloffame_filename="evolve_halloffame.pkl",
                 filename_prefix="",
                 n_halloffame=25):
        """
    Arguments
    ---------

      n_population: integer, optional, default=100
        Population size.

      n_generation: integer, optional, default=10
        Number of generations to for each repitition. Results are saved between
        repetitions.

      n_repetition: integer, optional, default=100
        Number of repititions to perform.

      n_features_init: int, optional, default=5
        Number of randomly selected features to initialize each individual with.

      pop_begin_filename: string, optional, default="population_begin.pkl"
        Filename where the initial population is read from, if it exists.

      pop_end_filename: string, optional, default="population_end.pkl"
        Filename where the final population is saved.

      halloffame_filename: string, optional, default="halloffame.pkl"
        Filename where a hall of fame is saved.

      filename_prefix: string
        Prefix for filenames, typically taken from input file filename excluding
        extension.

      n_halloffame: int, optional, default=25
        Number of individuals to save in the hall of fame

    """
        self.n_population = n_population
        self.n_generation = n_generation
        self.n_repetition = n_repetition

        self.n_features_init = n_features_init

        if len(filename_prefix) != 0:
            filename_prefix += "_"

        self.pop_begin_filename = filename_prefix + pop_begin_filename
        self.pop_end_filename = filename_prefix + pop_end_filename

        self.halloffame_filename = filename_prefix + halloffame_filename
        self.n_halloffame = n_halloffame


def default_stats(funcs=None):
    """
  Returns default deap.tools.Statistics object logging 'avg', 'std', 'min', 'max'.

  Arguments
  ---------

    funcs: dict, optional, default=(see below)
      A dict of functions to include in the deap.tools.Statistics object.

      Default:
        {
          "avg":np.mean,
          "std":np.std,
          "min":np.min,
          "max":np.max
        }


  Returns
  -------
    stats: deap.tools.Statistics
      The statistics to be logged.

  """
    if funcs is None:
        funcs = {"avg": np.mean, "std": np.std, "min": np.min, "max": np.max}
    stats = deap.tools.Statistics(key=lambda ind: ind.fitness.values)
    for key, val in six.iteritems(funcs):
        stats.register(key, val)
    return stats


class Log(object):
    """
  Attributes
  ----------

    stats: deap.tools.Statistics
        The statistics to be logged.

    logbook: deap.tools.Logbook

  """
    def __init__(self, stats=default_stats()):
        """
    Arguments
    ---------

      stats: deap.tools.Statistics, optional, default=default_stats()
        The statistics to be logged.
    """
        self.stats = stats
        self.logbook = deap.tools.Logbook()
        self.logbook.header = ['gen', 'nevals'
                               ] + (self.stats.fields if self.stats else [])

    def record(self, pop, gen, nevals, verbose=False):
        """
    Record stats for generation in logbook.

    Arguments
    ---------

      pop: iterable of individual
        Population to be minimized

      gen: integer
        Current generation index

      nevals: integer
        Number of times 'evaluate' has been called

      verbose: boolean
        Print information to stdout. If true, will print logbook.stream.
    """
        record = self.stats.compile(pop) if self.stats else {}
        self.logbook.record(gen=gen, nevals=nevals, **record)
        if verbose:
            print(self.logbook.stream)


def single_flip_children(parent):
    """
  Return all children of parent that differ by one feature on/off.

  Arguments
  ---------

    parent: List[bool] of length n_features
      This is a boolean list of shape [n_features], in which an element is True
      iff its corresponding feature is selected for retention.


  Returns
  -------

    offspring: iterable of individual
      Population to be evaluated

  """
    offspring = []
    for index in range(len(parent)):
        child = copy.deepcopy(parent)
        del child.fitness.values
        child[index] = not child[index]
        offspring.append(child)
    return offspring


def best_child(indiv, toolbox):
    """
  Return best child of a particular individual.

  Arguments
  ---------

    indiv: List[bool] of length n_features
      This is a boolean list of shape [n_features], in which an element is True
      iff its corresponding feature is selected for retention.

    toolbox: deap.base.Toolbox
      Contains methods used by deap during evolution. Expected to contain:

        toolbox.children(parent): Returns offspring (List[individual]) of parent

        toolbox.evaluate(indiv): To evaluate an individual's fitness

        toolbox.map(func, List[individual]): To map function executions

  Returns
  -------

    (best, nevals)

    best: List[bool] of length n_features
      The child with best fitness value

    nevals: integer
        Number of times 'evaluate' has been called
  """
    # generate children
    offspring = toolbox.children(indiv)
    nevals = evaluate_all(offspring, toolbox)
    return max(offspring, key=lambda child: child.fitness), nevals


def eaIndividualBestFirst(pop,
                          toolbox,
                          n_generation=10,
                          halloffame=None,
                          stats=default_stats(),
                          verbose=False):
    """
  Evolutionary algorithm that minimizes each individual in isolation by proposing
  children and replacing with the most fit child.

  Arguments
  ---------

    pop: iterable of individual
      Population to be minimized

    toolbox: deap.base.Toolbox
      Contains methods used during evolution. Expected to contain:

        toolbox.children(parent): Returns offspring (List[individual]) of parent

        toolbox.evaluate(indiv): To evaluate an individual's fitness

        toolbox.map(func, List[individual]): To map function executions

    n_generation: int, optional, default=10
      Number of generations to run.

    halloffame: deap.tools.HallOfFame, optional, default=None
      A Hall Of Fame containing the optimal solutions, as judged by CV score.

    stats: deap.tools.Statistics, optional, default=default_stats()
      The statistics to be logged.

    verbose: boolean
      Print information to stdout.


  Returns
  -------

    (pop, logbook)

    pop: iterable of individual
      The final population

    logbook: deap.tools.Logbook
      Logbook with the statistics of the evolution
  """
    remaining = len(pop)
    gen = 0

    # Evaluate & record in logbook
    nevals = evaluate_all(pop, toolbox)
    for indiv in pop:
        indiv.minimized = False

    log = Log(stats)
    log.record(pop, gen=gen, nevals=nevals, verbose=verbose)

    # initial halloffame update
    if halloffame is not None:
        halloffame.update(pop)

    while gen < n_generation:

        if remaining == 0:
            pop = initialize_population(len(pop), toolbox, verbose=verbose)
            remaining = len(pop)
            nevals = evaluate_all(pop, toolbox)
            for indiv in pop:
                indiv.minimized = False

        else:
            nevals_sum = 0
            # for each individual in population,
            for index, indiv in enumerate(pop):
                if not indiv.minimized:
                    best, nevals = best_child(indiv, toolbox)
                    nevals_sum += nevals

                    if best.fitness > indiv.fitness:
                        best.minimized = False
                        pop[index] = best
                    else:
                        indiv.minimized = True
                        remaining -= 1

        gen += 1

        # Update the (local) hall of fame with the generated individuals every
        # n_generation generations
        if halloffame is not None:
            halloffame.update(pop)

        # record in logbook
        log.record(pop, gen=gen, nevals=nevals_sum, verbose=verbose)

    return (pop, log.logbook)


def eaPopulationBestFirst(pop,
                          toolbox,
                          n_generation=10,
                          halloffame=None,
                          stats=default_stats(),
                          verbose=False):
    """
  Evolutionary algorithm that minimizes a population by selecting the most fit
  non-parent, generating children, and updating the population with the most
  fit individuals until the entire population has been a parent.

  Arguments
  ---------

    pop: iterable of individual
      Population to be minimized

    toolbox: deap.base.Toolbox
      Contains methods used during evolution. Expected to contain:

        toolbox.children(parent): Returns offspring (List[individual]) of parent

        toolbox.evaluate(indiv): To evaluate an individual's fitness

        toolbox.map(func, List[individual]): To map function executions

    n_generation: int, optional, default=10
      Number of generations between saving the hall of fame. Note, this only
      controls how often the hall of fame is saved. Minimization continues until
      all individuals in the population have had children.

    halloffame: deap.tools.HallOfFame, optional, default=None
      A Hall Of Fame containing the optimal solutions, as judged by CV score.

    stats: deap.tools.Statistics, optional, default=default_stats()
      The statistics to be logged.

    verbose: boolean
      Print information to stdout.


  Returns
  -------

    (pop, logbook)

    pop: iterable of individual
      The final population

    logbook: deap.tools.Logbook
      Logbook with the statistics of the evolution
  """

    gen = 0
    in_pop = pop

    # Evaluate & record in logbook
    nevals = evaluate_all(in_pop, toolbox)

    # use a HallOfFame for the population
    pop = casm.learn.create_halloffame(len(in_pop))
    pop.update(in_pop)

    # set as non-parents, if not specified
    for indiv in pop:
        if not hasattr(indiv, "parent"):
            indiv.parent = False

    # record initial population stats
    log = Log(stats)
    log.record(pop, gen=gen, nevals=nevals, verbose=verbose)

    while gen < n_generation:

        # get non-parents
        nonparents = [indiv for indiv in pop if not indiv.parent]

        if len(nonparents) == 0:
            new_pop = initialize_population(len(pop), toolbox, verbose=verbose)
            for indiv in new_pop:
                indiv.parent = False
            pop = casm.learn.create_halloffame(len(new_pop))
            pop.update(new_pop)
            nevals = evaluate_all(pop, toolbox)
            nonparents = [indiv for indiv in pop if not indiv.parent]

        # select most fit non-parent
        next_parent = max(nonparents, key=lambda indiv: indiv.fitness)
        next_parent.parent = True

        # generate children
        offspring = toolbox.children(next_parent)
        nevals = evaluate_all(offspring, toolbox)

        # set offspring as non-parents
        for indiv in offspring:
            indiv.parent = False

        # update population
        pop.update(offspring)

        gen += 1

        # Update the (local) hall of fame with the generated individuals every
        # n_generation generations
        if halloffame is not None:
            halloffame.update(pop)

        # record in logbook
        log.record(pop, gen=gen, nevals=nevals, verbose=verbose)

    return (pop, log.logbook)


class EvolutionaryFeatureSelection(BaseEstimator, SelectorMixin):
    def __init__(self,
                 algorithm,
                 estimator,
                 scoring=None,
                 cv=None,
                 penalty=0.0,
                 evolve_params_kwargs=dict(),
                 constraints_kwargs=dict(),
                 alg_args=list(),
                 alg_kwargs=dict(),
                 stats=default_stats(),
                 verbose=True):

        self.algorithm = algorithm
        self.alg_args = alg_args
        self.alg_kwargs = alg_kwargs

        self.estimator = estimator
        self.scoring = scoring
        self.cv = cv

        self.evolve_params = EvolutionaryParams(**evolve_params_kwargs)
        self.constraints = Constraints(**constraints_kwargs)

        self.penalty = penalty
        self.verbose = verbose

        self.toolbox = deap.base.Toolbox()
        self.stats = stats

    def _run(self):
        """
    Run the specified evolutionary algorithm.

    Arguments
    ---------

      algorithm: func,
        The evolutionary algorithm to perform

      alg_args: List[func]
        A list of functions used to generate positional arguments for 'algorithm'
        by calling f(self).

      alg_kwargs: dict
        A dict of key:function pairs used to generate keyword arguments for 'algorithm'.

          Example:

            alg_args = [
              operator.attrgetter("pop"),
              operatot.attrgetter("toolbox")]

            alg_kwargs = {
              "stats":operator.attrgetter("stats"),
              "halloffame":operatot.attrgetter("halloffame")
            }

          Then:

            in_args = [f(self) for f in alg_args]

            in_kwargs = dict()
            for key, f in six.iteritems(alg_kwargs):
              in_kwargs[key] = f(self)

            self.pop, self.logbook = algorithm(*in_args, **in_kwargs)


    Returns
    -------

      self: returns an instance of self.
    """
        ## read or construct hall of fame
        self.halloffame = initialize_halloffame(
            filename=self.evolve_params.halloffame_filename,
            n_halloffame=self.evolve_params.n_halloffame)

        ## read or construct initial population
        self.toolbox.decorate("population",
                              enforce_constraints(self.constraints))
        self.pop = initialize_population(
            self.evolve_params.n_population,
            self.toolbox,
            filename=self.evolve_params.pop_begin_filename,
            verbose=self.verbose)
        self.pop_begin = copy.deepcopy(self.pop)

        ## Run algorithm
        for rep in range(self.evolve_params.n_repetition):

            if self.verbose:
                print("Begin", rep + 1, "of", self.evolve_params.n_repetition,
                      "repetitions")

            if self.verbose:
                print("Begin", self.evolve_params.n_generation, "generations")
                t = time.perf_counter()

            in_args = [f(self) for f in self.alg_args]
            in_kwargs = dict()
            for key, f in six.iteritems(self.alg_kwargs):
                in_kwargs[key] = f(self)

            self.pop, self.logbook = self.algorithm(*in_args, **in_kwargs)
            self.pop_end = copy.deepcopy(self.pop)

            if self.verbose:
                print("Runtime:", time.perf_counter() - t, "(s)\n")

            ## Print end population
            if self.verbose:
                print("\nFinal population:")
                casm.learn.print_population(self.pop_end)

            save_population(self.pop,
                            filename=self.evolve_params.pop_end_filename,
                            verbose=self.verbose)

            ## Print hall of fame
            if self.verbose:
                print("\nHall of Fame:")
                casm.learn.print_population(self.halloffame)

            save_halloffame(self.halloffame,
                            filename=self.evolve_params.halloffame_filename,
                            verbose=self.verbose)

        return self

    def _get_support_mask(self):
        """
    Return most fit inidividual found.

    Returns
    -------

      support: List[bool] of length n_features
        This is a boolean list of shape [n_features], in which an element is True
        iff its corresponding feature is selected for retention.

    """
        return self.halloffame[0]

    def get_halloffame(self):
        """
    Returns the hall of fame of most fit individuals generated by 'fit'.

    Returns
    -------

      hall: deap.tools.HallOfFame
        A Hall Of Fame containing the optimal solutions, as judged by CV score.

    """
        return self.halloffame


class GeneticAlgorithm(EvolutionaryFeatureSelection):
    """
  Genetic algorithm for regression by optimizing a CV score.

  Implements deap.algorithms.eaSimple, using selTournament, for selection,
  cxUniform for mating, and mutFlipBit for mutation. The probability of mating
  and mutating is set to 1.0.

  The population of solutions is saved

  For more details see: http://deap.gel.ulaval.ca/doc/0.8/api/algo.html#complete-algorithms

  Attributes
  ---------

    estimator:  estimator object implementing 'fit'
      The estimator specified by the input settings.

    scoring: string
      A string or a scorer callable object / function with signature
      scorer(estimator, X, y). The parameter for sklearn.model_selection.cross_val_score,
      default = None, uses estimator.score().

    cv: cross-validation generator or an iterable
      Provides train/test splits

    penalty: float
      The CV score is increased by 'penalty*(number of selected basis function)'

    evolve_params: casm.learn.evolve.EvolutionaryParams
      A EvolutionaryParams object containing parameters.

    constraints: casm.learn.evolve.Constraints
      A Constraints object containing constraints on allowed individuals.

    selTournamentSize: int
      Tournament size. A larger tournament size weeds out less fit
      individuals more quickly, while a smaller tournament size weeds out
      less fit individuals more gradually.

    cxUniformProb: float
      Probability of swapping bits during mating.

    mutFlipBitProb: float
      Probability of mutating bits "constraints".

    CrossOverProb: 1.0
      Probability of performing mating.

    MutateProb: 1.9
      Probability of performing mutation.

    toolbox: deap.base.Toolbox
      Contains methods used by deap during evolution.

    verbose: boolean
      Print information to stdout.

    hall: deap.tools.HallOfFame
      A Hall Of Fame containing the optimal solutions, as judged by CV score.

    pop_begin: iterable of individuals
      The initial population of individual solutions.

    pop_end: iterable of individuals
      The final population of individual solutions.



  """
    def __init__(self,
                 estimator,
                 scoring=None,
                 cv=None,
                 penalty=0.0,
                 evolve_params_kwargs=dict(),
                 constraints_kwargs=dict(),
                 selTournamentSize=3,
                 cxUniformProb=0.5,
                 mutFlipBitProb=0.01,
                 verbose=True):
        """
    Arguments
    ---------

      estimator:  estimator object implementing 'fit'
        The estimator specified by the input settings.

      scoring: string, callable or None, optional, default=None
        A string or a scorer callable object / function with signature
        scorer(estimator, X, y). The parameter is passed to sklearn.model_selection.cross_val_score,
        has a default=None which uses estimator.score().

      cv: cross-validation generator or an iterable, optional, default=None
        Provides train/test splits. The parameter is passed to sklearn.model_selection.cross_val_score,
        has a default=None which uses KFold cross-validation with k=3.

      penalty: float, optional, default=0.0
        The CV score is increased by 'penalty*(number of selected basis function)'

      evolve_params_kwargs: dict, optional, default=dict()
        Keyword arguments for construction of EvolutionaryParams object containing
        parameters. See casm.learn.evolve.EvolutionaryParams for possibilities.

      constraints_kwargs: dict, optional, default=dict()
        Keyword arguments for construction of Constraints object containing
        constraints on allowed individuals. See casm.learn.evolve.Constraints for
        possibilities.

      selTournamentSize: int, optional, default=3
        Tournament size. A larger tournament size weeds out less fit
        individuals more quickly, while a smaller tournament size weeds out
        less fit individuals more gradually.

      cxUniformProb: float, optional, default=0.5
        Probability of swapping bits during mating.

      mutFlipBitProb: float, optional, default=0.01
        Probability of mutating bits "constraints".

      verbose: boolean, optional, default=True
        Print information to stdout.

    """
        super(GeneticAlgorithm,
              self).__init__(deap.algorithms.eaSimple,
                             estimator,
                             scoring=scoring,
                             cv=cv,
                             penalty=penalty,
                             evolve_params_kwargs=evolve_params_kwargs,
                             constraints_kwargs=constraints_kwargs,
                             verbose=verbose,
                             alg_args=[
                                 attrgetter("pop"),
                                 attrgetter("toolbox"),
                                 attrgetter("CrossOverProb"),
                                 attrgetter("MutateProb"),
                                 lambda obj: obj.evolve_params.n_generation
                             ],
                             alg_kwargs={
                                 "verbose": attrgetter("verbose"),
                                 "halloffame": attrgetter("halloffame"),
                                 "stats": attrgetter("stats")
                             })

        self.selTournamentSize = selTournamentSize
        self.cxUniformProb = cxUniformProb
        self.mutFlipBitProb = mutFlipBitProb

        # currently fixed:
        self.CrossOverProb = 1.0
        self.MutateProb = 1.0

        ## Selection
        self.toolbox.register("select",
                              deap.tools.selTournament,
                              tournsize=self.selTournamentSize)

        ## Crossover
        self.toolbox.register("mate",
                              deap.tools.cxUniform,
                              indpb=self.cxUniformProb)
        self.toolbox.decorate("mate", enforce_constraints(self.constraints))

        ## Mutation
        self.toolbox.register("mutate",
                              deap.tools.mutFlipBit,
                              indpb=self.mutFlipBitProb)
        self.toolbox.decorate("mutate", enforce_constraints(self.constraints))

    def fit(self, X, y):
        """
    Run genetic algorithm to generate optimal solutions for X*b = y.

    The best solutions generated are stored in self.halloffame.

    Arguments
    ---------

      X: array-like of shape (n_samples, n_features)
        The input data

      y: array-like of shape (n_samples, 1)
        The values


    Returns
    -------

      self: returns an instance of self.
    """

        ## Register toolbox functions that depend on X,y
        self.toolbox.register("individual", initNRandomOn,
                              casm.learn.creator.Individual, X.shape[1],
                              self.evolve_params.n_features_init)
        self.toolbox.register("population", deap.tools.initRepeat, list,
                              self.toolbox.individual)
        self.toolbox.register("evaluate",
                              casm.learn.model_selection.cross_val_score,
                              self.estimator,
                              X,
                              y=y,
                              scoring=self.scoring,
                              cv=self.cv,
                              penalty=self.penalty)

        return self._run()


class IndividualBestFirst(EvolutionaryFeatureSelection):
    """
  Individual best first algorithm for regression by optimizing a CV score.

  Implements a best first search optimization for each individual in the initial
  population. Each individual in the population is minimized by repeatedly begin
  replaced by its most fit child.

  By default, children are generated by generating all the individual that differ
  from the parent by +/- 1 selected feature. But any generating function may be
  given that generates offspring from a single parent individual.


  Attributes
  ---------

    estimator:  estimator object implementing 'fit'
      The estimator specified by the input settings.

    scoring: string
      A string or a scorer callable object / function with signature
      scorer(estimator, X, y). The parameter for sklearn.model_selection.cross_val_score,
      default = None, uses estimator.score().

    cv: cross-validation generator or an iterable
      Provides train/test splits

    penalty: float
      The CV score is increased by 'penalty*(number of selected basis function)'

    evolve_params: casm.learn.evolve.EvolutionaryParams
      A EvolutionaryParams object containing parameters.

    constraints: casm.learn.evolve.Constraints
      A Constraints object containing constraints on allowed individuals.

    children: func
      A function with signature 'offspring = func(indiv)'.

    toolbox: deap.base.Toolbox
      Contains methods used by deap during evolution.

    verbose: boolean
      Print information to stdout.

    hall: deap.tools.HallOfFame
      A Hall Of Fame containing the optimal solutions, as judged by CV score.

    pop_begin: iterable of individuals
      The initial population of individual solutions.

    pop_end: iterable of individuals
      The final population of individual solutions.
  """
    def __init__(self,
                 estimator,
                 scoring=None,
                 cv=None,
                 penalty=0.0,
                 evolve_params_kwargs=dict(),
                 constraints_kwargs=dict(),
                 children=single_flip_children,
                 verbose=True):
        """
    Arguments
    ---------

      estimator:  estimator object implementing 'fit'
        The estimator specified by the input settings.

      scoring: string, callable or None, optional, default=None
        A string or a scorer callable object / function with signature
        scorer(estimator, X, y). The parameter is passed to sklearn.model_selection.cross_val_score,
        has a default=None which uses estimator.score().

      cv: cross-validation generator or an iterable, optional, default=None
        Provides train/test splits. The parameter is passed to sklearn.model_selection.cross_val_score,
        has a default=None which uses KFold cross-validation with k=3.

      penalty: float, optional, default=0.0
        The CV score is increased by 'penalty*(number of selected basis function)'

      evolve_params_kwargs: dict, optional, default=dict()
        Keyword arguments for construction of EvolutionaryParams object containing
        parameters. See casm.learn.evolve.EvolutionaryParams for possibilities.

      constraints_kwargs: dict, optional, default=dict()
        Keyword arguments for construction of Constraints object containing
        constraints on allowed individuals. See casm.learn.evolve.Constraints for
        possibilities.

      children: func, optional, default=single_flip_children
        A function with signature 'offspring = func(indiv)'.

      verbose: boolean, optional, default=True
        Print information to stdout.

    """

        # add a count of the number of individuals fully optimized
        stats = default_stats()
        stats_indiv = deap.tools.Statistics(key=lambda indiv: indiv.minimized)
        stats_indiv.register("completed", sum)
        mstats = deap.tools.MultiStatistics(fitness=stats,
                                            optimization=stats_indiv)

        super(IndividualBestFirst,
              self).__init__(eaIndividualBestFirst,
                             estimator,
                             scoring=scoring,
                             cv=cv,
                             penalty=penalty,
                             evolve_params_kwargs=evolve_params_kwargs,
                             constraints_kwargs=constraints_kwargs,
                             verbose=verbose,
                             alg_args=[
                                 attrgetter("pop"),
                                 attrgetter("toolbox"),
                                 lambda obj: obj.evolve_params.n_generation
                             ],
                             alg_kwargs={
                                 "verbose": attrgetter("verbose"),
                                 "halloffame": attrgetter("halloffame"),
                                 "stats": attrgetter("stats")
                             },
                             stats=mstats)

        self.children = children

        self.toolbox.register("children", self.children)
        self.toolbox.decorate("children",
                              restrict_constraints(self.constraints))

    def fit(self, X, y):
        """
    Run individual best first algorithm to generate optimal solutions for X*b = y.

    The best solutions generated are stored in self.halloffame.

    Arguments
    ---------

      X: array-like of shape (n_samples, n_features)
        The input data

      y: array-like of shape (n_samples, 1)
        The values


    Returns
    -------

      self: returns an instance of self.
    """
        ## Register toolbox functions that depend on X,y
        self.toolbox.register("individual", initNRandomOn,
                              casm.learn.creator.Individual, X.shape[1],
                              self.evolve_params.n_features_init)
        self.toolbox.register("population", deap.tools.initRepeat, list,
                              self.toolbox.individual)
        self.toolbox.register("evaluate",
                              casm.learn.model_selection.cross_val_score,
                              self.estimator,
                              X,
                              y=y,
                              scoring=self.scoring,
                              cv=self.cv,
                              penalty=self.penalty)

        return self._run()


class PopulationBestFirst(EvolutionaryFeatureSelection):
    """
  Population best first algorithm for regression by optimizing a CV score.

  Implements a best first search optimization for a population of individual
  solutions. Each individual is associated with a 'status' that indicates
  whether it has had children yet or not. At each step, the most fit individual
  that hasn't had children has children and the population is updated to keep
  only the 'n_population' most fit individuals. The algorithm stops when all
  individuals in the population have had children.

  By default, children are generated by generating all the individual that differ
  from the parent by +/- 1 selected feature. But any generating function may be
  given that generates offspring from a single parent individual.


  Attributes
  ----------

    estimator:  estimator object implementing 'fit'
      The estimator specified by the input settings.

    scoring: string
      A string or a scorer callable object / function with signature
      scorer(estimator, X, y). The parameter for sklearn.model_selection.cross_val_score,
      default = None, uses estimator.score().

    cv: cross-validation generator or an iterable
      Provides train/test splits

    penalty: float
      The CV score is increased by 'penalty*(number of selected basis function)'

    evolve_params: casm.learn.evolve.EvolutionaryParams
      A EvolutionaryParams object containing parameters.

    constraints: casm.learn.evolve.Constraints
      A Constraints object containing constraints on allowed individuals.

    children: func
      A function with signature 'offspring = func(indiv)'.

    toolbox: deap.base.Toolbox
      Contains methods used by deap during evolution.

    verbose: boolean
      Print information to stdout.

    hall: deap.tools.HallOfFame
      A Hall Of Fame containing the optimal solutions, as judged by CV score.

    pop_begin: iterable of individuals
      The initial population of individual solutions.

    pop_end: iterable of individuals
      The final population of individual solutions.
  """
    def __init__(self,
                 estimator,
                 scoring=None,
                 cv=None,
                 penalty=0.0,
                 evolve_params_kwargs=dict(),
                 constraints_kwargs=dict(),
                 children=single_flip_children,
                 verbose=True):
        """
    Arguments
    ---------

      estimator:  estimator object implementing 'fit'
        The estimator specified by the input settings.

      scoring: string, callable or None, optional, default=None
        A string or a scorer callable object / function with signature
        scorer(estimator, X, y). The parameter is passed to sklearn.model_selection.cross_val_score,
        has a default=None which uses estimator.score().

      cv: cross-validation generator or an iterable, optional, default=None
        Provides train/test splits. The parameter is passed to sklearn.model_selection.cross_val_score,
        has a default=None which uses KFold cross-validation with k=3.

      penalty: float, optional, default=0.0
        The CV score is increased by 'penalty*(number of selected basis function)'

      evolve_params_kwargs: dict, optional, default=dict()
        Keyword arguments for construction of EvolutionaryParams object containing
        parameters. See casm.learn.evolve.EvolutionaryParams for possibilities.

      constraints_kwargs: dict, optional, default=dict()
        Keyword arguments for construction of Constraints object containing
        constraints on allowed individuals. See casm.learn.evolve.Constraints for
        possibilities.

      children: func, optional, default=single_flip_children
        A function with signature 'offspring = func(indiv)'.

      verbose: boolean, optional, default=True
        Print information to stdout.

    """
        # add a count of the number of individuals fully optimized
        stats = default_stats()
        stats_indiv = deap.tools.Statistics(key=lambda indiv: indiv.parent)
        stats_indiv.register("parents", sum)
        mstats = deap.tools.MultiStatistics(fitness=stats,
                                            optimization=stats_indiv)

        super(PopulationBestFirst,
              self).__init__(eaPopulationBestFirst,
                             estimator,
                             scoring=scoring,
                             cv=cv,
                             penalty=penalty,
                             evolve_params_kwargs=evolve_params_kwargs,
                             constraints_kwargs=constraints_kwargs,
                             verbose=verbose,
                             alg_args=[
                                 attrgetter("pop"),
                                 attrgetter("toolbox"),
                                 lambda obj: obj.evolve_params.n_generation
                             ],
                             alg_kwargs={
                                 "verbose": attrgetter("verbose"),
                                 "halloffame": attrgetter("halloffame"),
                                 "stats": attrgetter("stats")
                             },
                             stats=mstats)

        self.children = children

        self.toolbox.register("children", self.children)
        self.toolbox.decorate("children",
                              restrict_constraints(self.constraints))

    def fit(self, X, y):
        """
    Run population best first algorithm to generate optimal solutions for X*b = y.

    The best solutions generated are stored in self.halloffame.

    Arguments
    ---------

      X: array-like of shape (n_samples, n_features)
        The input data

      y: array-like of shape (n_samples, 1)
        The values


    Returns
    -------

      self: returns an instance of self.
    """
        ## Register toolbox functions that depend on X,y
        self.toolbox.register("individual", initNRandomOn,
                              casm.learn.creator.Individual, X.shape[1],
                              self.evolve_params.n_features_init)
        self.toolbox.register("population", deap.tools.initRepeat, list,
                              self.toolbox.individual)
        self.toolbox.register("evaluate",
                              casm.learn.model_selection.cross_val_score,
                              self.estimator,
                              X,
                              y=y,
                              scoring=self.scoring,
                              cv=self.cv,
                              penalty=self.penalty)

        return self._run()



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import numpy as np
from casm.vasp.io import poscar, kpoints, species, incar


class VaspIO:
    """ Generate a set of VASP input files from settings files

        Contains:
            self.incar: Incar object
            self.poscar: Poscar object
            self.kpoints: Kpoints object
            self.species: Species dict
    """
    def __init__(self,
                 incarfile,
                 ref_kpointsfile,
                 ref_structurefile,
                 structurefile,
                 speciesfile,
                 sort=False):
        """ Construct a VaspIO object

            Args:
                incarfile:  path to INCAR file
                ref_kpointsfile: path to a reference KPOINTS file
                ref_structurefile: path to a reference CASM structure.json or VASP POSCAR file, used for scaling k-points
                structurefile: path to CASM structure.json or VASP POSCAR file
                    for the configuration
                speciesfile: path to SPECIES file

            This functions reads the input files and generates self.kpoints appropriate for self.poscar given that 'ref_kpointsfile' is for 'ref_structurefile'.
        """
        ref_structure = poscar.Poscar(ref_structurefile)
        ref_kpoints = kpoints.Kpoints(ref_kpointsfile)

        self.species = species.species_settings(speciesfile)
        self.poscar = poscar.Poscar(structurefile, self.species)
        self.incar = incar.Incar(incarfile, self.species, self.poscar, sort)
        self.kpoints = ref_kpoints.super_kpoints(ref_structure, self.poscar)

    def write_potcar(self, filename, sort=False):
        """ Write an appropriate POTCAR """
        if sort == False:
            try:
                file = open(filename, 'w')
            except:
                raise VaspIOError("Could not open file for writing: '" +
                                  filename + "'")
            for name in self.poscar.type_atoms:
                potcar = open(
                    os.path.join(self.species[name].potcardir, 'POTCAR'))
                file.write(potcar.read())
                potcar.close()
            file.close()
        else:

            if True:
                # new code

                # dict: key = alias, value = list of Sites
                pos = self.poscar.basis_dict()

                with open(filename, 'w') as file:
                    # for each alias
                    for alias in sorted(pos.keys()):
                        # find matching IndividualSpecies with write_potcar == True
                        for name in self.species:
                            if self.species[
                                    name].alias == alias and self.species[
                                        name].write_potcar:
                                # add to POTCAR file
                                with open(
                                        os.path.join(
                                            self.species[name].potcardir,
                                            'POTCAR')) as potcar:
                                    file.write(potcar.read())
                                break

            else:
                # original code:

                # dict: key = alias, value = list of Sites
                pos = self.poscar.basis_dict()

                # list of unique basis occupants
                basis = []

                # dict: key = alias, value = list of Site occupants (? why not just use pos ?)
                alias = dict()
                for k in sorted(pos.keys()):
                    alias[k] = []
                    for atom in pos[k]:
                        if atom.occupant not in basis:
                            basis.append(atom.occupant)
                        if atom.occupant not in alias[k]:
                            alias[k].append(atom.occupant)

                # list of IndividualSpecies names
                specie = [species[x].name for x in self.species]

                # list of potcardir
                potcar_list = []

                # loop over unique basis occupants
                for s in basis:
                    try:
                        if (self.species[s].write_potcar):
                            potcar_list.append(self.species[s].potcardir)
                        else:
                            # ? don't understand this ?
                            if len(alias[self.species[s].alias]) <= 1:
                                potcar_list.append(self.species[s].potcardir)
                    except ValueError:
                        continue

                # write POTCAR
                with open(filename, 'w') as outFile:
                    for f in potcar_list:
                        with open(os.path.join(f, 'POTCAR')) as inFile:
                            outFile.write(inFile.read())

    def write(self, dirpath):
        """ Write VASP input files in directory 'dirpath' """
        self.poscar.write(os.path.join(dirpath, 'POSCAR'), self.sort)
        self.incar.write(os.path.join(dirpath, 'INCAR'))
        self.kpoints.write(os.path.join(dirpath, 'KPOINTS'))
        self.write_potcar(os.path.join(dirpath, 'POTCAR'), self.sort)



"""implements the parent class for vasp calculations"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import json
import math
import numpy as np
import os
import pandas
import sys

try:
    from prisms_jobs import Job, JobDB, error_job, complete_job, JobsError, JobDBError, EligibilityError
except ImportError:
    # use of the pbs module is deprecated after CASM v0.2.1
    from pbs import Job, JobDB, error_job, complete_job, JobDBError, EligibilityError
    from pbs import PBSError as JobsError

from casm.project import Project, Selection, attribute_info
from casm.vasp.io import attribute_classes
from casm import vasp
from casm.misc import noindent
from casm.vaspwrapper import VaspWrapperError, read_settings, write_settings, \
    vasp_input_file_names


class VaspCalculatorBase(object):
    """
    Base class containing all the basic functions that method classes can inherit

    Attributes
    ----------
    selection : casm.project.Selection
        selection of configuration
    calctype : string
        calctype to setup and run the neb calculations
    auto : bool
    sort : bool

    """
    def __init__(self, selection, calctype=None, auto=True, sort=True):
        """set up attributes for the base class"""
        self.selection = selection
        self.calctype = calctype
        self.auto = auto
        self.sort = sort
        self.casm_directories = self.selection.proj.dir
        self.casm_settings = self.selection.proj.settings
        if self.casm_settings is None:
            raise VaspWrapperError(
                "Not in a CASM project. The file '.casm' directory was not found."
            )

        self.clex = self.casm_settings.default_clex
        if calctype:
            self.clex.calctype = calctype
        self.calc_subdir = ""
        # everything between $(calcdir)/run.*/ and OSZICAR and OUTCAR files
        self.results_subdir = ''
        self.calculator = None
        self.append_selection_data()

    @classmethod
    def from_configuration_dir(cls,
                               configuration_dir,
                               calctype,
                               auto=True,
                               sort=True):
        """returns a instance of the Neb class instantited with a single configuration"""
        # change config_dir to configuration_dir all over
        proj = Project(configuration_dir)
        sel = Selection(proj, "EMPTY", "config", False)
        split_path = configuration_dir.split(os.path.sep)
        index = split_path.index("training_data")
        name = '/'.join(split_path[index + 1:])
        sel.data = pandas.DataFrame({
            "name": name,
            "selected": 1
        },
                                    index=range(1))
        # try:
        #     os.mkdir(os.path.join(proj.path, ".casm/tmp"))
        # except:
        #     pass
        # sel_config = sel.saveas(os.path.join(proj.path, ".casm/tmp", configname.replace('/', '.')), True)
        obj = cls(sel, calctype, auto, sort)
        return obj

    def config_properties(self, config_data):
        """read properties directories of a specific configuration"""
        config_dict = dict(config_data)
        config_dict["configdir"] = self.casm_directories.configuration_dir(
            config_data["name"], self.calc_subdir)
        config_dict["calcdir"] = self.casm_directories.calctype_dir(
            config_data["name"], self.clex, self.calc_subdir)
        config_dict["setfile"] = self.casm_directories.settings_path_crawl(
            "calc.json", config_data["name"], self.clex, self.calc_subdir)
        return config_dict

    def append_selection_data(self):
        """append configproperties to selection.data"""
        config_dicts = []
        for index, config_data in self.selection.data.iterrows():
            config_dicts.append(self.config_properties(config_data))
        properites_dict = {}
        for key in config_dicts[0].keys():
            if key in self.selection.data.columns:
                continue
            properites_dict[key] = []
            for config_dict in config_dicts:
                properites_dict[key].append([config_dict[key]])
            self.selection.add_data(key, properites_dict[key])

    def pre_setup(self):
        """has to be overloaded in the child class"""
        pass

    def setup(self):
        """Setup initial relaxation run for the selection"""
        self.pre_setup()
        for index, config_data in self.selection.data.iterrows():
            self.config_setup(config_data)

    def config_setup(self, config_data):
        """ Setup initial relaxation run

            Uses the following files from the most local .../settings/calctype.name directory:
                INCAR: VASP input settings
                KPOINTS: VASP kpoints settings
                POSCAR: reference for KPOINTS if KPOINTS mode is not A/AUTO/Automatic
                SPECIES: info for each species such as which POTCAR files to use, MAGMOM, GGA+U, etc.

            Uses the following files from the .../config_calcdir/00 directory:
                POSCAR: sample structure of the configuration to be relaxed

        """
        settings = self.read_settings(config_data["setfile"])
        vaspfiles = self.get_vasp_input_files(config_data, settings)
        incarfile, ref_kpointsfile, ref_structurefile, structurefile, speciesfile, extra_input_files = vaspfiles
        vasp.io.write_vasp_input(config_data["calcdir"], incarfile,
                                 ref_kpointsfile, ref_structurefile,
                                 structurefile, speciesfile, self.sort,
                                 extra_input_files, settings["strict_kpoints"])

    def get_deformation(self, settings):
        """ Either reads or queries for deformation matrix from settings dict"""
        if settings["initial_deformation"]["method"] == 'manual':
            deformation = np.array(
                settings["initial_deformation"]["deformation"])
        elif settings["initial_deformation"]["method"] == 'auto':
            configname = settings["initial_deformation"]["configname"]
            calctype = settings["initial_deformation"]["calctype"]
            sel_tmp = Selection(self.selection.proj, "EMPTY", "config", False)
            sel_tmp.data = pandas.DataFrame(
                {
                    "configname": configname,
                    "selected": 1
                }, index=range(1))
            try:
                os.mkdir(os.path.join(self.selection.proj.path, ".casm/tmp"))
            except:
                pass
            sel_config = sel_tmp.saveas(
                os.path.join(self.selection.proj.path, ".casm/tmp",
                             configname.replace('/', '.')), True)
            sel_config.query(["relaxation_strain(U,0:5,{})".format(calctype)])
            deformation = np.array([
                float(sel_config.data["relaxation_strain(U,{},{})".format(
                    i, calctype)].loc[0]) for i in range(6)
            ])
            if os.path.isfile(
                    os.path.join(self.selection.proj.path, ".casm/tmp",
                                 configname.replace('/', '.'))):
                os.remove(
                    os.path.join(self.selection.proj.path, ".casm/tmp",
                                 configname.replace('/', '.')))
        else:
            raise VaspWrapperError(
                "use manual or auto mode to set initial deformation. see casm format --vasp for settings"
            )

        if deformation.ndim == 1:
            deformation = np.array(
                [[deformation[0], deformation[5], deformation[4]],
                 [deformation[5], deformation[1], deformation[3]],
                 [deformation[4], deformation[3], deformation[2]]])
        return deformation

    def apply_deformation(self, deformation, calcdir):
        """applies the deformation and write the poscar to calcdir"""
        poscarfile = os.path.join(calcdir, "POSCAR")
        poscar_obj = vasp.io.poscar.Poscar(poscarfile)
        poscar_obj.apply_deformation(deformation)
        poscar_obj.write(poscarfile)

    @staticmethod
    def read_settings(setfile):
        """ Read settings from a settings calc.json file"""

        settings = read_settings(setfile)
        # set default settings if not present
        if not "ncore" in settings:
            settings["ncore"] = None
        if not "npar" in settings:
            settings["npar"] = None
        if not "kpar" in settings:
            settings["kpar"] = None
        if not "vasp_cmd" in settings:
            settings["vasp_cmd"] = None
        if not "ncpus" in settings:
            settings["ncpus"] = None
        if not "run_limit" in settings:
            settings["run_limit"] = None
        if not "prerun" in settings:
            settings["prerun"] = None
        if not "postrun" in settings:
            settings["postrun"] = None

        return settings

    def get_vasp_input_files(self, config_data, settings):
        # Find required input files in CASM project directory tree
        vaspfiles = vasp_input_file_names(self.casm_directories,
                                          config_data["name"], self.clex,
                                          self.calc_subdir)
        incarfile, ref_kpointsfile, ref_structurefile, structurefile, speciesfile = vaspfiles
        # Find optional input files
        extra_input_files = []
        for s in settings["extra_input_files"]:
            extra_input_files.append(
                self.casm_directories.settings_path_crawl(
                    s, config_data["name"], self.clex, self.calc_subdir))
            if extra_input_files[-1] is None:
                raise vasp.VaspError("Neb.setup failed. Extra input file " +
                                     s + " not found in CASM project.")
        if settings["initial"]:
            extra_input_files += [
                self.casm_directories.settings_path_crawl(
                    settings["initial"], config_data["name"], self.clex,
                    self.calc_subdir)
            ]
            if extra_input_files[-1] is None:
                raise vasp.VaspError(
                    "Neb.setup failed. No initial INCAR file " +
                    settings["initial"] + " found in CASM project.")
        if settings["final"]:
            extra_input_files += [
                self.casm_directories.settings_path_crawl(
                    settings["final"], config_data["name"], self.clex,
                    self.calc_subdir)
            ]
            if extra_input_files[-1] is None:
                raise vasp.VaspError("Neb.setup failed. No final INCAR file " +
                                     settings["final"] +
                                     " found in CASM project.")
        return incarfile, ref_kpointsfile, ref_structurefile, structurefile, speciesfile, extra_input_files

    def is_converged(self, calculation):
        # Check for electronic convergence in completed calculations. Returns True or False.
        # Verify that the last relaxation reached electronic convergence
        for i in range(len(calculation.rundir)):
            try:
                print(calculation.rundir[-i - 1])
                vrun_oszicar = vasp.io.Oszicar(
                    os.path.join(calculation.calcdir,
                                 calculation.rundir[-i - 1],
                                 self.results_subdir, "OSZICAR"))
                vrun_nelm = vasp.io.get_incar_tag(
                    "NELM",
                    os.path.join(calculation.calcdir,
                                 calculation.rundir[-i - 1]))
                if vrun_nelm == None:
                    vrun_nelm = 60  # pushing the default. may be write an addon to get it from outcar
                if len(vrun_oszicar.num_elm[-1]) >= vrun_nelm:
                    print(
                        'The last relaxation run (' +
                        os.path.basename(relaxation.rundir[-i - 1]) +
                        ') failed to achieve electronic convergence; properties.calc.json will not be written.\n'
                    )
                    self.report_status(calculation.calcdir, 'failed',
                                       'electronic_convergence')
                    return False
                break
            except:
                pass

        # Verify that the final static run reached electronic convergence
        vrun_oszicar = vasp.io.Oszicar(
            os.path.join(calculation.calcdir, "run.final", self.results_subdir,
                         "OSZICAR"))
        vrun_nelm = vasp.io.get_incar_tag(
            "NELM", os.path.join(calculation.calcdir, "run.final"))
        if vrun_nelm == None:
            vrun_nelm = 60  # pushing the default. may be write an addon to get it from outcar
        if vrun_oszicar.num_elm[-1] >= vrun_nelm:
            print(
                'The final run failed to achieve electronic convergence; properties.calc.json will not be written.\n'
            )
            self.report_status(calculation.calcdir, 'failed',
                               'electronic_convergence')
            return False

        return True

    def submit(self):
        """ submit jobs for a selection"""
        self.pre_setup()
        db = JobDB()
        for index, config_data in self.selection.data.iterrows():
            print("Submitting...")
            print("Configuration:", config_data["name"])
            #first, check if the job has already been submitted and is not completed
            print("Calculation directory:", config_data["calcdir"])
            id = db.select_regex_id("rundir", config_data["calcdir"])
            print("JobID:", id)
            sys.stdout.flush()
            try:
                if id != []:
                    db.update()
                    for j in id:
                        job = db.select_job(j)
                        if job["jobstatus"] != "C":
                            print("JobID:", job["jobid"], "  Jobstatus:",
                                  job["jobstatus"], "  Not submitting.")
                            sys.stdout.flush()
                            raise BreakException
            except BreakException:
                continue
            settings = self.read_settings(config_data["setfile"])
            # construct the Relax object
            calculation = self.calculator(config_data["calcdir"],
                                          self.run_settings(settings))
            # check the current status
            (status, task) = calculation.status()

            if status == "complete":
                print("Status:", status, "  Not submitting.")
                sys.stdout.flush()

                # ensure job marked as complete in db
                if self.auto:
                    for j in id:
                        job = db.select_job(j)
                        if job["taskstatus"] == "Incomplete":
                            try:
                                complete_job(jobid=j)
                            except (JobsError, JobDBError,
                                    EligibilityError) as e:
                                print(str(e))
                                sys.stdout.flush()

                # ensure results report written
                if not os.path.isfile(
                        os.path.join(config_data["calcdir"],
                                     "properties.calc.json")):
                    if (is_converged(calculation)):
                        self.finalize(config_data)

                continue

            elif status == "not_converging":
                print("Status:", status, "  Not submitting.")
                sys.stdout.flush()
                continue

            elif status != "incomplete":
                raise VaspWrapperError("unexpected relaxation status: '" +
                                       status + "' and task: '" + task + "'")
                sys.stdout.flush()
                continue

            print("Preparing to submit a VASP relaxation PBS job")
            sys.stdout.flush()

            # cd to configdir, submit jobs from configdir, then cd back to currdir
            currdir = os.getcwd()
            os.chdir(config_data["calcdir"])

            self.config_setup(config_data)
            nodes, ppn = self._calc_submit_node_info(settings, config_data)

            # construct command to be run
            cmd = ""
            if settings["preamble"] is not None:
                # Append any instructions given in the 'preamble' file, if given
                preamble = self.casm_directories.settings_path_crawl(
                    settings["preamble"], config_data["name"], self.clex,
                    self.calc_subdir)
                with open(preamble) as my_preamble:
                    cmd += "".join(my_preamble)
            # Or just execute a single prerun line, if given
            if settings["prerun"] is not None:
                cmd += settings["prerun"] + "\n"
            cmd += self.run_cmd(config_data["configdir"], self.calctype)
            if settings["postrun"] is not None:
                cmd += settings["postrun"] + "\n"

            print("Constructing a PBS job")
            sys.stdout.flush()
            # construct a Job
            job = Job(name=casm.jobname(config_data["configdir"]),
                      account=settings["account"],
                      nodes=nodes,
                      ppn=ppn,
                      walltime=settings["walltime"],
                      pmem=settings["pmem"],
                      qos=settings["qos"],
                      queue=settings["queue"],
                      message=settings["message"],
                      email=settings["email"],
                      priority=settings["priority"],
                      command=cmd,
                      auto=self.auto,
                      software=db.config["software"])

            print("Submitting")
            sys.stdout.flush()
            # submit the job
            job.submit()
            self.report_status(config_data["calcdir"], "submitted")

            # return to current directory
            os.chdir(currdir)

            print("CASM VASPWrapper relaxation PBS job submission complete\n")
            sys.stdout.flush()

    def run_cmd(self, configdir, calctype):
        """has to be overloaded in the method class"""
        return None

    @staticmethod
    def _calc_submit_node_info(settings, config_data):
        """return nodes, ppn from settings of a configuration"""
        if settings["nodes"] != None and settings["ppn"] != None:
            return int(settings["nodes"]), int(settings["ppn"])
        elif settings["atom_per_proc"] != None and settings["ppn"] != None:
            pos = vasp.io.Poscar(os.path.join(config_data["calcdir"],
                                              "POSCAR"))
            num = len(pos.basis)
            nodes = int(
                math.ceil(
                    float(num) / float(settings["atom_per_proc"]) /
                    float(settings["ppn"])))
            return nodes, int(settings["ppn"])
        elif settings["nodes_per_image"] != None and settings["ppn"] != None:
            nodes = int(config_data["n_images"]) * \
                float(settings["nodes_per_image"])
            return nodes, int(settings["ppn"])
        else:
            raise VaspWrapperError(
                "Not enough information to determine nodes and ppn information"
            )

    @staticmethod
    def run_settings(settings):
        """ Set default values based on runtime environment"""

        # set default values
        if settings["npar"] == "CASM_DEFAULT":
            if "PBS_NUM_NODES" in os.environ:
                settings["npar"] = int(os.environ["PBS_NUM_NODES"])
            elif "SLURM_JOB_NUM_NODES" in os.environ:
                settings["npar"] = int(os.environ["SLURM_JOB_NUM_NODES"])
            else:
                settings["npar"] = None
        elif settings["npar"] == "VASP_DEFAULT":
            settings["npar"] = None

        if settings["npar"] is None:
            if settings["ncore"] == "CASM_DEFAULT":
                if "PBS_NUM_PPN" in os.environ:
                    settings["ncore"] = int(os.environ["PBS_NUM_PPN"])
                elif "SLURM_CPUS_ON_NODE" in os.environ:
                    settings["ncore"] = int(os.environ["SLURM_CPUS_ON_NODE"])
                else:
                    settings["ncore"] = None
            elif settings["ncore"] == "VASP_DEFAULT":
                settings["ncore"] = 1
        else:
            settings["ncore"] = None

        if settings["ncpus"] is None or settings["ncpus"] == "CASM_DEFAULT":
            if "PBS_NP" in os.environ:
                settings["ncpus"] = int(os.environ["PBS_NP"])
            elif "SLURM_NTASKS" in os.environ:
                settings["ncpus"] = int(os.environ["SLURM_NTASKS"])
            else:
                settings["ncpus"] = None

        if settings["run_limit"] is None or settings[
                "run_limit"] == "CASM_DEFAULT":
            settings["run_limit"] = 10

        return settings

    def run(self):
        """run the job of a selection"""
        for index, config_data in self.selection.data.iterrows():
            settings = self.read_settings(config_data["setfile"])
            calculation = self.calculator(config_data["calcdir"],
                                          self.run_settings(settings))

            # check the current status
            (status, task) = calculation.status()

            if status == "complete":
                print("Status:", status)
                sys.stdout.flush()

                # mark job as complete in db
                if self.auto:
                    try:
                        complete_job()
                    except (JobsError, JobDBError, EligibilityError) as e:
                        print(str(e))
                        sys.stdout.flush()

                # write results to properties.calc.json
                if (is_converged(calculation)):
                    self.finalize(config_data)
                continue

            elif status == "not_converging":
                print("Status:", status)
                self.report_status(config_data["calcdir"], "failed",
                                   "run_limit")
                print("Returning")
                sys.stdout.flush()
                continue

            elif status == "incomplete":

                self.report_status(config_data["calcdir"], "started")
                (status, task) = calculation.run()

            else:
                self.report_status(config_data["calcdir"], "failed", "unknown")
                raise VaspWrapperError("unexpected relaxation status: '" +
                                       status + "' and task: '" + task + "'")
            sys.stdout.flush()

            # once the run is done, update database records accordingly

            if status == "not_converging":

                # mark error
                if self.auto:
                    try:
                        error_job("Not converging")
                    except (JobsError, JobDBError) as e:
                        print(str(e))
                        sys.stdout.flush()

                print("Not Converging!")
                sys.stdout.flush()
                self.report_status(config_data["calcdir"], "failed",
                                   "run_limit")

                # print a local settings file, so that the run_limit can be extended if the
                #   convergence problems are fixed

                config_set_dir = self.casm_directories.configuration_calc_settings_dir(
                    config_data["name"], self.clex, self.calc_subdir)

                try:
                    os.makedirs(config_set_dir)
                except:
                    pass
                settingsfile = os.path.join(config_set_dir, "calc.json")
                write_settings(settings, settingsfile)

                print("Writing:", settingsfile)
                print("Edit the 'run_limit' property if you wish to continue.")
                sys.stdout.flush()
                continue

            elif status == "complete":

                # mark job as complete in db
                if self.auto:
                    try:
                        complete_job()
                    except (JobsError, JobDBError, EligibilityError) as e:
                        print(str(e))
                        sys.stdout.flush()

                # write results to properties.calc.json
                if is_converged(calculation):
                    self.finalize(config_data)

            else:
                self.report_status(config_data["calcdir"], "failed", "unknown")
                raise VaspWrapperError(
                    "vasp relaxation complete with unexpected status: '" +
                    status + "' and task: '" + task + "'")
            sys.stdout.flush()

    def report_status(self, calcdir, status, failure_type=None):
        """Report calculation status to status.json file in configuration directory.

            Args:
                status: string describing calculation status. Currently used values are
                    not_submitted
                    submitted
                    complete
                    failed
                failure_type: optional string describing reason for failure. Currently used values are
                    unknown
                    electronic_convergence
                    run_limit"""

        output = dict()
        output["status"] = status
        if failure_type is not None:
            output["failure_type"] = failure_type

        outputfile = os.path.join(calcdir, "status.json")
        with open(outputfile, 'w') as file:
            file.write(
                json.dumps(output,
                           cls=noindent.NoIndentEncoder,
                           indent=4,
                           sort_keys=True))
        print("Wrote " + outputfile)
        sys.stdout.flush()

    def report(self):
        """
        report status for the selection
        Notes
        -----
        checks for convergence
        calls the finalize function to write the approprite properties files.
        """
        for index, config_data in self.selection.data.iterrows():
            try:
                settings = self.read_settings(config_data["setfile"])
                calculation = self.calculator(config_data["calcdir"],
                                              self.run_settings(settings))
                if self.is_converged(calculation):
                    self.finalize(config_data)
            except:
                print(
                    "Unable to report properties for directory {}.\n"
                    "Please verify that it contains a completed VASP calculation."
                    .format(config_data["configdir"]))
                raise

    def finalize(self, config_data, structurefile=None):
        # write properties.calc.json
        vaspdir = os.path.join(config_data["calcdir"], "run.final")
        speciesfile = self.casm_directories.settings_path_crawl(
            "SPECIES", config_data["name"], self.clex, self.calc_subdir)
        #Making structurefile from a structure.json file
        structurefile = self.casm_directories.structure_json(
            config_data["name"])
        output = self.properties(vaspdir, structurefile, speciesfile)

        outputfile = os.path.join(config_data["calcdir"],
                                  "properties.calc.json")
        with open(outputfile, 'w') as file:
            file.write(
                json.dumps(output,
                           cls=noindent.NoIndentEncoder,
                           indent=4,
                           sort_keys=True))
        print("Wrote " + outputfile)
        sys.stdout.flush()
        self.report_status(config_data["calcdir"], 'complete')

    @staticmethod
    def properties(vaspdir, initial_structurefile=None, speciesfile=None):
        """ return a dict of output form a vasp directory"""
        dof_info = attribute_info.AttributeInfo(initial_structurefile)
        output = dict()
        # load the OSZICAR and OUTCAR
        zcar = vasp.io.Oszicar(os.path.join(vaspdir, "OSZICAR"))
        ocar = vasp.io.Outcar(os.path.join(vaspdir, "OUTCAR"))

        # the calculation is run on the 'sorted' POSCAR, need to report results 'unsorted'

        if (initial_structurefile is not None) and (speciesfile is not None):
            species_settings = vasp.io.species_settings(speciesfile)
            initial_structure = vasp.io.Poscar(initial_structurefile,
                                               species_settings)
            unsort_dict = initial_structure.unsort_dict()
        else:
            # fake unsort_dict (unsort_dict[i] == i)
            initial_structure = vasp.io.Poscar(os.path.join(vaspdir, "POSCAR"))
            unsort_dict = dict(
                zip(range(0, len(structure.basis)),
                    range(0, len(structure.basis))))
        contcar = vasp.io.Poscar(os.path.join(vaspdir, "CONTCAR"))

        # unsort_dict:
        #   Returns 'unsort_dict', for which: unsorted_dict[orig_index] == sorted_index;
        #   unsorted_dict[sorted_index] == orig_index
        #   For example:
        #     'unsort_dict[0]' returns the index into the unsorted POSCAR of the first atom in the sorted POSCAR
        output["atom_type"] = initial_structure.atom_type
        #output["atoms_per_type"] = initial_structure.num_atoms
        output["coord_mode"] = contcar.coord_mode

        # as lists
        output["lattice"] = [
            noindent.NoIndent(list(v)) for v in contcar.lattice()
        ]
        output["atom_coords"] = [None for i in range(len(contcar.basis))]
        for i, ba in enumerate(contcar.basis):
            output["atom_coords"][unsort_dict[i]] = noindent.NoIndent(
                list(ba.position))

        output["atom_dofs"] = {}
        output["atom_dofs"]["force"] = {}
        output["atom_dofs"]["force"]["value"] = [
            None for i in range(len(ocar.forces))
        ]
        for i, force in enumerate(ocar.forces):
            output["atom_dofs"]["force"]["value"][
                unsort_dict[i]] = noindent.NoIndent(force)

        output["global_dofs"] = {}
        output["global_dofs"]["energy"] = {}
        output["global_dofs"]["energy"]["value"] = zcar.E[-1]

        if dof_info.atom_dofs is not None:
            if "Cmagspin" in list(dof_info.atom_dofs.keys()):
                output["global_dofs"]["Cmagspin"] = {}
                cmagspin_specific_output = attribute_classes.CmagspinAttr(
                    dof_info).vasp_output_dictionary(ocar)
                output["atom_dofs"].update(cmagspin_specific_output)
                #TODO: Need a better way to write global magmom. I don't like what I did here
                output["global_dofs"]["Cmagspin"]["value"] = zcar.mag[-1]

            #TODO: When you don't have Cmagspin but have magnetic calculations. This part can be removed if you runall magnetic calculations as Cmagspin calculations.
            #TODO: Need a better way of doing this. Some code duplication here.
            else:
                if ocar.ispin == 2:
                    output["global_dofs"]["Cmagspin"] = {}
                    output["global_dofs"]["Cmagspin"]["value"] = zcar.mag[-1]
                    if ocar.lorbit in [1, 2, 11, 12]:
                        output["atom_dofs"]["Cmagspin"] = {}
                        output["atom_dofs"]["Cmagspin"]["value"] = [
                            None for i in range(len(contcar.basis))
                        ]

                        for i, v in enumerate(contcar.basis):
                            output["atom_dofs"]["Cmagspin"]["value"][
                                unsort_dict[i]] = [
                                    noindent.NoIndent(ocar.mag[i])
                                ]

        #TODO: Code duplication here. If you have a magnetic calculation without dofs, you still need to write magmom values. This can be removed if you run all the magnetic calculations as Cmagspin dof calculations.
        #TODO: If you still want to have this particular functionality, wrap it up in a helper function to avoid code duplication.
        else:
            if ocar.ispin == 2:
                output["global_dofs"]["Cmagspin"]["value"] = zcar.mag[-1]
                if ocar.lorbit in [1, 2, 11, 12]:
                    output["atom_dofs"]["Cmagspin"] = {}
                    output["atom_dofs"]["Cmagspin"]["value"] = [
                        None for i in range(len(contcar.basis))
                    ]

                    for i, v in enumerate(contcar.basis):
                        output["atom_dofs"]["Cmagspin"]["value"][
                            unsort_dict[i]] = [noindent.NoIndent(ocar.mag[i])]
        return output


class BreakException(Exception):
    """use this exception to break an outer loop"""
    pass



"""See class info"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

### External ###
import os
import shutil
import six
from math import ceil, sqrt
import sys
import json
# import re
# import warnings
try:
    from prisms_jobs import Job, JobDB, error_job, complete_job, JobsError, JobDBError, EligibilityError
except ImportError:
    # use of the pbs module is deprecated after CASM v0.2.1
    from pbs import Job, JobDB, error_job, complete_job, JobDBError, EligibilityError
    from pbs import PBSError as JobsError

### Local ###
from casm import vasp, wrapper
from casm.misc import noindent
from casm.project import DirectoryStructure, ProjectSettings
from casm.vaspwrapper import VaspWrapperError, read_settings, write_settings, \
  vasp_input_file_names

### Globals ###
VALID_PROP_TYPES = ["KPOINTS", "ENCUT", "NBANDS", "SIGMA"]
VALID_TOL_TYPES = ["relaxed_energy", "relaxed_volume", "relaxed_lattice"]

#pylint: disable=line-too-long, too-many-instance-attributes, too-many-arguments, too-many-branches, too-many-statements, too-many-locals, too-many-nested-blocks, too-many-lines


class ConvergeError(Exception):
    """ Exception sub-class for errors related to convergence """
    pass


class Converge(object):
    """The Converge class contains functions for setting up, executing, and parsing a VASP convergence test.

        The convergence creates the following directory structure:
        config/
            calctype.name/
                property/
                    value.0/
                    value.1/
                    ...
                    value.n/

        'value.i' directories are created for each discrete step in 'value'
          being explored for 'property', e.g., every 10 eV for ENCUT, or every
          1 for KPOINTS.
        If runs are static runs, 'value.i' contains INCAR, POSCAR, etc.
        If runs are relaxation runs, 'value.i' contains 'run.0', 'run.1', ...,
          'run.final', where 'run.final' contains a static run.

        This automatically looks for VASP settings files using:
          casm.project.DirectoryStructure.settings_path_crawl

    Attributes
    ----------

      casm_settings: casm.project.ProjectSettings instance
        CASM project settings

      casm_directories: casm.project.DirectoryStructure instance
        CASM project directory hierarchy

      settings: dict
        Settings for job submission and the relaxation, see vaspwrapper.read_settings

      configdir: str
        Directory where configuration results are stored. The result of:
          casm.project.DirectoryStructure.configuration_dir(self.configname)

      configname: str
        The name of the configuration to be calculated

      auto: boolean
        True if using prisms_jobs module's JobDB to manage jobs

      sort: boolean
        True if sorting atoms in POSCAR by type

      clex: casm.project.ClexDescription instance
        The cluster expansion being worked on. Used for the 'calctype' settings.
        Currently, fixed to self.casm_settings.default_clex.

    """
    def __init__(self,
                 configdir=None,
                 auto=True,
                 sort=True,
                 propdir=None,
                 prop=None):
        """
        Construct a VASP convergence job object.

        Arguments
        ----------

            configdir: str, optional, default=None
              Path to configuration directory. If None, uses the current working
              directory
            auto: boolean, optional, default=True,
              Use True to use the prisms_jobs module's JobDB to manage jobs

            sort: boolean, optional, default=True,
              Use True to sort atoms in POSCAR by type

            propdir: str, optional, default=None
              Name of the directory within the configdir for converging the
                current property-of-interest

            prop: str, optional, default=None
              Name of an INCAR property that is being converged

        """
        print("Construct a casm.vaspwrapper.Converge instance:")

        if configdir is None:
            configdir = os.getcwd()
        print("  Input directory:", configdir)

        self.propdir = propdir
        self.prop = prop

        # get the configname from the configdir path
        _res = os.path.split(configdir)
        self.configname = os.path.split(_res[0])[1] + "/" + _res[1]
        print("  Configuration:", self.configname)

        print("Reading CASM settings")
        self.casm_directories = DirectoryStructure(configdir)
        self.casm_settings = ProjectSettings(configdir)
        if self.casm_settings is None:
            raise VaspWrapperError(
                "Not in a CASM project. The file '.casm' directory was not found."
            )

        if os.path.abspath(configdir) != self.configdir:
            print("")
            print("input configdir:", configdir)
            print("determined configname:", self.configname)
            print("expected configdir given configname:", self.configdir)
            raise VaspWrapperError("Mismatch between configname and configdir")

        # fixed to default_clex for now
        self.clex = self.casm_settings.default_clex

        # store path to .../config/calctype.name, and create if not existing
        self.calcdir = self.casm_directories.calctype_dir(
            self.configname, self.clex)
        try:
            os.mkdir(self.calcdir)
        except:
            pass
        print("  Calculations directory:", self.calcdir)

        # read the settings json file
        print("  Reading converge.json settings file")
        sys.stdout.flush()
        setfile = self.casm_directories.settings_path_crawl(
            "converge.json", self.configname, self.clex)

        if setfile is None:
            raise VaspWrapperError(
                "Could not find \"converge.json\" in an appropriate \"settings\" directory"
            )

        else:
            print("  Read settings from:", setfile)
        self.settings = read_settings(setfile)

        # add required keys to settings if not present
        if not "ncore" in self.settings:
            self.settings["ncore"] = None
        if not "npar" in self.settings:
            self.settings["npar"] = None
        if not "kpar" in self.settings:
            self.settings["kpar"] = None
        if not "vasp_cmd" in self.settings:
            self.settings["vasp_cmd"] = None
        if not "ncpus" in self.settings:
            self.settings["ncpus"] = None
        if not "run_limit" in self.settings:
            self.settings["run_limit"] = None
        if not "prerun" in self.settings:
            self.settings["prerun"] = None
        if not "postrun" in self.settings:
            self.settings["postrun"] = None
        if not "prop" in self.settings:
            self.settings["prop"] = None
        if not "prop_start" in self.settings:
            self.settings["prop_start"] = None
        if not "prop_step" in self.settings:
            self.settings["prop_step"] = None
        if not "prop_stop" in self.settings:
            self.settings["prop_stop"] = None
        if not "tol" in self.settings:
            self.settings["tol"] = "relaxed_energy"
        if not "tol_amount" in self.settings:
            self.settings["tol_amount"] = 0.001

        self.auto = auto
        self.sort = sort

        # Invoke the settings checker
        self._settings_checker()

        # Check if a name was specified
        if self.settings["name"] is None:
            self.name = self.settings["prop"] + "_converge"
        else:
            self.name = self.settings["name"]

        # If we have a list-type prop_start (e.g. a k-point string), we need to be careful with making our list
        if isinstance(self.settings["prop_start"], list):
            # For constructing directory names, we'll want a 3-position sub
            self.prop_name = "%i_%i_%i"
            self.prop_list = []
            i, j, k = self.settings["prop_start"]
            while [i, j, k] <= self.settings["prop_stop"]:
                self.prop_list += [[i, j, k]]
                i += self.settings["prop_step"][0]
                j += self.settings["prop_step"][1]
                k += self.settings["prop_step"][2]
        # Anything integer or float, however, is easily range'd
        elif float(self.settings["prop_start"]) != int(
                self.settings["prop_start"]) or float(
                    self.settings["prop_stop"]) != int(
                        self.settings["prop_stop"]) or float(
                            self.settings["prop_step"]) != int(
                                self.settings["prop_step"]):
            # For constructing directory names, we'll want a 1-position sub
            self.prop_name = "%g"
            self.prop_list = list(
                self.frange(
                    self.settings["prop_start"],
                    self.settings["prop_stop"] + self.settings["prop_step"],
                    self.settings["prop_step"]))
        else:
            # For constructing directory names, we'll want a 1-position sub
            self.prop_name = "%i"
            self.prop_list = range(
                self.settings["prop_start"],
                self.settings["prop_stop"] + self.settings["prop_step"],
                self.settings["prop_step"])

        # Making all the directories
        self.prop_dir_list = []
        for prop in self.prop_list:
            if isinstance(prop, list):
                self.prop_dir_list += [
                    os.path.join(self.calcdir, self.prop_name % tuple(prop))
                ]
            else:
                self.prop_dir_list += [
                    os.path.join(self.calcdir, self.prop_name % prop)
                ]
            try:
                if isinstance(prop, list):
                    os.mkdir(
                        os.path.join(self.calcdir,
                                     self.prop_name % tuple(prop)))
                else:
                    os.mkdir(os.path.join(self.calcdir, self.prop_name % prop))
            except OSError:
                pass

        print("  DONE\n")
        sys.stdout.flush()

    def collect(self):
        """ Collect the results of a convergence run """

        nrg_data = []
        vol_data = []
        lat_a_data = []
        lat_b_data = []
        lat_c_data = []
        # Iterate over each individual value of the property, since parallelism!
        for prop, propdir in zip(self.prop_list, self.prop_dir_list):
            # construct the Converge object
            convergence = vasp.Converge(propdir, self.run_settings(), prop)

            # check the current status
            (status, task) = convergence.status()  #pylint: disable=unused-variable

            # If any job isn't done, we should abort
            if status == "incomplete":
                print("Directory " + propdir + " Has Status:" + status +
                      "  Please wait for the job to finish.")
                return
            elif status != "complete":
                print("Directory " + propdir + " Has Status:" + status +
                      "  Please inspect for errors, then re-submit.")
                return
            else:
                # ensure results report written
                if not os.path.isfile(
                        os.path.join(propdir, "properties.calc.json")):
                    ###TODO: Why did I write it like this? Why not call convergence.finalize()?
                    store_prop = self.prop
                    store_propdir = self.propdir

                    self.prop = prop
                    self.propdir = propdir

                    self.finalize()

                    self.prop = store_prop
                    self.propdir = store_propdir

                # Collect job data
                my_properties = read_properties(
                    os.path.join(propdir, "properties.calc.json"))

                nrg_data += [
                    float(my_properties["relaxed_energy"]) /
                    sum(my_properties["atoms_per_type"])
                ]  #Energy per atom!!!
                vol_data += [self.volume(my_properties["relaxed_lattice"])]
                lat_a, lat_b, lat_c = self.lengths(
                    my_properties["relaxed_lattice"])  #pylint: disable=invalid-name
                lat_a_data += [lat_a]
                lat_b_data += [lat_b]
                lat_c_data += [lat_c]

        # Calculate the diffs
        nrg_diff = []
        vol_diff = []
        lat_a_diff = []
        lat_b_diff = []
        lat_c_diff = []
        prior_nrg = 0.
        prior_vol = 0.
        prior_lat_a = 0.
        prior_lat_b = 0.
        prior_lat_c = 0.
        for nrg, vol, lat_a, lat_b, lat_c in zip(nrg_data, vol_data,
                                                 lat_a_data, lat_b_data,
                                                 lat_c_data):
            nrg_diff += [nrg - prior_nrg]
            vol_diff += [vol - prior_vol]
            lat_a_diff += [lat_a - prior_lat_a]
            lat_b_diff += [lat_b - prior_lat_b]
            lat_c_diff += [lat_c - prior_lat_c]
            prior_nrg = nrg
            prior_vol = vol
            prior_lat_a = lat_a
            prior_lat_b = lat_b
            prior_lat_c = lat_c
        abs_nrg_diff = [abs(x) for x in nrg_diff]
        abs_vol_diff = [abs(x) for x in vol_diff]
        abs_lat_a_diff = [abs(x) for x in lat_a_diff]
        abs_lat_b_diff = [abs(x) for x in lat_b_diff]
        abs_lat_c_diff = [abs(x) for x in lat_c_diff]

        abs_nrg_diff_per = [
            100. * abs(x / y) for x, y in zip(abs_nrg_diff, nrg_data)
        ]
        abs_vol_diff_per = [
            100. * abs(x / y) for x, y in zip(abs_vol_diff, vol_data)
        ]
        abs_lat_a_diff_per = [
            100. * abs(x / y) for x, y in zip(abs_lat_a_diff, lat_a_data)
        ]
        abs_lat_b_diff_per = [
            100. * abs(x / y) for x, y in zip(abs_lat_b_diff, lat_b_data)
        ]
        abs_lat_c_diff_per = [
            100. * abs(x / y) for x, y in zip(abs_lat_c_diff, lat_c_data)
        ]

        # Determine where, if anywhere, we have reached convergence
        prop_conv = None
        prop_idx = -1
        if self.settings["tol"] is None:
            pass
        else:
            # Enumerate over our specified tolerances
            for tol, tol_amount in zip(self.settings["tol"],
                                       self.settings["tol_amount"]):
                if tol.lower() == "relaxed_energy":

                    # If even the last step doesn't hit tol, we have NOT converged, and should abort this loop now!!!
                    if abs_nrg_diff[-1] > tol_amount:
                        prop_idx = -1
                        break

                    # Walk forward in steps and see when we've converged
                    for idx, nrg_diff_i in enumerate(abs_nrg_diff):

                        # Check if THIS value fulfills the tol (the value may oscillate, so this is NOT a break condition)
                        if nrg_diff_i <= tol_amount:
                            continue

                        # Increment our "global" prop index, as appropriate
                        else:
                            if idx > prop_idx:
                                prop_idx = idx

                    continue

                elif tol.lower() == "relaxed_volume":

                    # If even the last step doesn't hit tol, we have NOT converged, and should abort this loop now!!!
                    if abs_vol_diff[-1] > tol_amount:
                        prop_idx = -1
                        break

                    # Walk forward in steps and see when we've converged
                    for idx, vol_diff_i in enumerate(abs_vol_diff):

                        # Check if THIS value fulfills the tol (the value may oscillate, so this is NOT a break condition)
                        if vol_diff_i <= tol_amount:
                            continue

                        # Increment our "global" prop index, as appropriate
                        else:
                            if idx > prop_idx:
                                prop_idx = idx

                    continue

                elif tol.lower() == "relaxed_lattice":
                    # If even the last step doesn't hit tol, we have NOT converged, and should abort this loop now!!!
                    if abs_lat_a_diff[-1] > tol_amount or abs_lat_b_diff[
                            -1] > tol_amount or abs_lat_c_diff[-1] > tol_amount:
                        prop_idx = -1
                        break

                    # Walk forward in steps and see when we've converged
                    for idx, (lat_a_diff_i, lat_b_diff_i,
                              lat_c_diff_i) in enumerate(
                                  zip(abs_lat_a_diff, abs_lat_b_diff,
                                      abs_lat_c_diff)):

                        # Check if THIS value fulfills the tol (the value may oscillate, so this is NOT a break condition)
                        if lat_a_diff_i <= tol_amount and lat_b_diff_i <= tol_amount and lat_c_diff_i <= tol_amount:
                            continue

                        # Increment our "global" prop index, as appropriate
                        else:
                            if idx > prop_idx:
                                prop_idx = idx

                    continue

                else:
                    raise ConvergeError("Invalid tolerance type %s!" % tol)

            if prop_idx >= 0:
                # We want to do prop_idx + 1 because we stop incrementing the idx counter on the step BEFORE
                # the tol, and we want the step AT the tol
                prop_conv = self.prop_list[prop_idx + 1]

        # Print our results to convergence.calc.json
        conv_dict = {}
        conv_dict["property"] = {
            "property_name": self.settings["prop"],
            "property_converged_value": prop_conv
        }
        # conv_dict["property"] = self.settings["prop"]
        # conv_dict["property_converged_value"] = prop_conv
        conv_dict["tolerance"] = {
            "tolerance_type": self.settings["tol"],
            "tolerance_value": self.settings["tol_amount"]
        }
        # conv_dict["tolerances"] = self.settings["tol"]
        # conv_dict["tolerance_amounts"] = self.settings["tol_amount"]
        conv_dict["data"] = {
            "property_values": self.prop_list,
            "energy": {
                "relaxed": nrg_data,
                "abs_relaxed_diff": abs_nrg_diff,
                "percent_relaxed_diff": abs_nrg_diff_per
            },
            "volume": {
                "relaxed": vol_data,
                "abs_relaxed_diff": abs_vol_diff,
                "percent_relaxed_diff": abs_vol_diff_per
            },
            "lattice_vector": {
                "relaxed":
                [[a, b, c]
                 for a, b, c in zip(lat_a_data, lat_b_data, lat_c_data)],
                "abs_relaxed_diff": [[a, b, c] for a, b, c in zip(
                    abs_lat_a_diff, abs_lat_b_diff, abs_lat_c_diff)],
                "percent_relaxed_diff":
                [[a, b, c]
                 for a, b, c in zip(abs_lat_a_diff_per, abs_lat_b_diff_per,
                                    abs_lat_c_diff_per)]
            }
        }

        # conv_dict["values"] = self.prop_list
        # conv_dict["relaxed_energies"] = nrg_data
        # conv_dict["relaxed_energy_diffs"] = nrg_diff
        # conv_dict["abs_relaxed_energy_diffs"] = abs_nrg_diff
        # conv_dict["relaxed_volumes"] = vol_data
        # conv_dict["relaxed_volume_diffs"] = vol_diff
        # conv_dict["abs_relaxed_volume_diffs"] = abs_vol_diff
        # conv_dict["relaxed_lattice_vector_lengths"] = [[a, b, c] for a, b, c in zip(lat_a_data, lat_b_data, lat_c_data)]
        # conv_dict["relaxed_lattice_vector_length_diffs"] = [[a, b, c] for a, b, c in zip(lat_a_diff, lat_b_diff, lat_c_diff)]
        # conv_dict["abs_relaxed_lattice_vector_length_diffs"] = [[a, b, c] for a, b, c in zip(abs_lat_a_diff, abs_lat_b_diff, abs_lat_c_diff)]
        # write properties.calc.json

        conv_file_name = "convergence.calc.json"
        outputfile = os.path.join(self.calcdir, conv_file_name)
        with open(outputfile, 'wb') as my_file:
            my_file.write(
                six.u(
                    json.dumps(conv_dict,
                               cls=noindent.NoIndentEncoder,
                               indent=4,
                               sort_keys=True)).encode('utf-8'))
        print("Wrote " + outputfile)
        sys.stdout.flush()

        # Copy our converge.json so we can revisit it later
        write_settings(self.settings,
                       os.path.join(self.calcdir, "converge.json"))
        print("Wrote " + os.path.join(self.calcdir, "converge.json"))
        sys.stdout.flush()

    def setup(self):
        """ Setup initial convergence run

            Uses the following files from the most local .../settings/calctype.name directory:
                INCAR: VASP input settings
                KPOINTS: VASP kpoints settings
                structure.json: reference for KPOINTS if KPOINTS mode is not A/AUTO/Automatic
                SPECIES: info for each species such as which POTCAR files to use, MAGMOM, GGA+U, etc.

            Uses the following files from the .../config directory:
                POS: structure of the configuration to be converged

        """
        # Find required input files in CASM project directory tree
        vaspfiles = vasp_input_file_names(self.casm_directories,
                                          self.configname, self.clex)
        incarfile, kpointsfile, _, structurefile, speciesfile = vaspfiles

        # Verify that required input files exist
        if incarfile is None:
            raise vasp.VaspError(
                "Converge.setup failed. No INCAR file found in CASM project.")
        if kpointsfile is None:
            raise vasp.VaspError(
                "Converge.setup failed. No KPOINTS file found in CASM project."
            )
        if structurefile is None:
            raise vasp.VaspError(
                "Converge.setup failed. No POS file found for this configuration."
            )
        if speciesfile is None:
            raise vasp.VaspError(
                "Converge.setup failed. No SPECIES file found in CASM project."
            )

        # Find optional input files
        extra_input_files = []
        for s in self.settings["extra_input_files"]:
            extra_input_files.append(
                self.casm_directories.settings_path_crawl(
                    s, self.configname, self.clex))
            if extra_input_files[-1] is None:
                raise vasp.VaspError(
                    "Converge.setup failed. Extra input file " + s +
                    " not found in CASM project.")
        if self.settings["initial"]:
            extra_input_files += [
                self.casm_directories.settings_path_crawl(
                    self.settings["initial"], self.configname, self.clex)
            ]
            if extra_input_files[-1] is None:
                raise vasp.VaspError(
                    "Converge.setup failed. No initial INCAR file " +
                    self.settings["initial"] + " found in CASM project.")
        if self.settings["final"]:
            extra_input_files += [
                self.casm_directories.settings_path_crawl(
                    self.settings["final"], self.configname, self.clex)
            ]
            if extra_input_files[-1] is None:
                raise vasp.VaspError(
                    "Converge.setup failed. No final INCAR file " +
                    self.settings["final"] + " found in CASM project.")
        sys.stdout.flush()

        print("Setting up VASP input files:", self.propdir)

        # read kpoints
        print("  Reading reference KPOINTS:", kpointsfile)
        kpoints = vasp.io.Kpoints(kpointsfile)

        # read species, structure, incar, and generate kpoints
        print("  Reading SPECIES:", speciesfile)
        species_settings = vasp.io.species_settings(speciesfile)
        print("  Reading POS/structure.json:", structurefile)
        poscar = vasp.io.Poscar(structurefile, species_settings)
        print("  Reading INCAR:", incarfile)
        incar = vasp.io.Incar(incarfile, species_settings, poscar, True)

        # Modify as appropriate
        if self.settings["prop"].upper() == "ENCUT":
            try:
                incar.tags["ENCUT"] = int(self.prop)
            except:
                raise ConvergeError(
                    "Error in Cconverge.collect: something has gone wrong and the run-specific property %s could not be cast as int!"
                    % self.prop)
        elif self.settings["prop"].upper() == "NBANDS":
            try:
                incar.tags["NBANDS"] = int(self.prop)
            except:
                raise ConvergeError(
                    "Error in Converge.collect: something has gone wrong and the run-specific property %s could not be cast as int!"
                    % self.prop)
        elif self.settings["prop"].upper() == "SIGMA":
            try:
                incar.tags["SIGMA"] = float(self.prop)
            except:
                raise ConvergeError(
                    "Error in Converge.collect: something has gone wrong and the run-specific property %s could not be cast as int!"
                    % self.prop)
        elif self.settings["prop"].upper() == "KPOINTS":
            if isinstance(self.prop, list):
                if kpoints.automode[0].lower() == "a":
                    raise ConvergeError(
                        "Error in Converge.collect: An \"Automatic\" k-point generation scheme was supplied in %s, but you gave me 3 k-points values: %i, %i, %i!"
                        % (kpointsfile, self.prop[0], self.prop[1],
                           self.prop[2]))
                kpoints.subdivisions = self.prop
            else:
                if kpoints.automode[0].lower() == "a":
                    kpoints.subdivisions = self.prop
                else:
                    kpoints.subdivisions = [self.prop, self.prop, self.prop]

        else:
            raise ConvergeError(
                "Error in Converge.collect: \"prop: %s\" not a valid convergence prop type!\nCurrently supported convergence prop types are %s"
                % (self.settings["prop"], VALID_PROP_TYPES))

        # write main input files
        print("  Writing POSCAR:", os.path.join(self.propdir, 'POSCAR'))
        poscar.write(os.path.join(self.propdir, 'POSCAR'), True)
        print("  Writing INCAR:", os.path.join(self.propdir, 'INCAR'))
        incar.write(os.path.join(self.propdir, 'INCAR'))
        print("  Writing KPOINTS:", os.path.join(self.propdir, 'KPOINTS'))
        kpoints.write(os.path.join(self.propdir, 'KPOINTS'))
        print("  Writing POTCAR:", os.path.join(self.propdir, 'POTCAR'))
        vasp.io.write_potcar(os.path.join(self.propdir, 'POTCAR'), poscar,
                             species_settings, True)

        # copy extra input files
        print("  Copying extra input files", end=' ')
        for my_input_file in extra_input_files:
            print(my_input_file, end=' ')
            shutil.copy(my_input_file, self.propdir)
        print("")

        print("VASP input files complete\n")
        sys.stdout.flush()

    def submit(self):
        """Submit jobs for this VASP convergence"""

        print("Submitting...")
        print("Configuration:", self.configname)
        print("Property:", self.settings["prop"])
        # Iterate over each individual value of the property, since parallelism!
        for prop, propdir in zip(self.prop_list, self.prop_dir_list):
            try:
                # Store our propdir before editing it
                store_propdir = self.propdir
                store_prop = self.prop
                self.propdir = propdir
                self.prop = prop
                print("Value:", self.prop)
                # first, check if the job has already been submitted and is not completed
                db = JobDB()  #pylint: disable=invalid-name
                print("Calculation directory:", propdir)
                jobid = db.select_regex_id("rundir", propdir)
                print("JobID:", jobid)
                sys.stdout.flush()
                if jobid != []:
                    for j in jobid:
                        job = db.select_job(j)
                        # taskstatus = ["Incomplete", "Complete", "Continued", "Check", "Error:.*", "Aborted"]
                        # jobstatus = ["C", "Q", "R", "E", "W", "H", "M"]
                        if job["jobstatus"] != "C":
                            print("JobID:", job["jobid"], "  Jobstatus:",
                                  job["jobstatus"], "  Not submitting.")
                            sys.stdout.flush()
                            raise UserWarning()
                        #elif job["taskstatus"] in ["Complete", "Check"] or re.match("Error:.*", job["taskstatus"]):
                        #    print "JobID:", job["jobid"], "  Taskstatus:", job["taskstatus"], "  Not submitting."
                        #    sys.stdout.flush()
                        #    return

                # second, only submit a job if convergence status is "incomplete"

                # construct the Converge object
                convergence = vasp.Converge(propdir, self.run_settings(), prop)
                # convergence = vasp.Converge(self.calcdir, self.run_settings())

                # check the current status
                (status, task) = convergence.status()

                if status == "complete":
                    print("Status:", status, "  Not submitting.")
                    sys.stdout.flush()

                    # ensure job marked as complete in db
                    if self.auto:
                        for j in jobid:
                            job = db.select_job(j)
                            if job["taskstatus"] == "Incomplete":
                                try:
                                    complete_job(jobid=j)
                                except (JobsError, JobDBError,
                                        EligibilityError) as my_error:
                                    print(str(my_error))
                                    sys.stdout.flush()

                    # ensure results report written
                    if not os.path.isfile(
                            os.path.join(propdir, "properties.calc.json")):
                        self.finalize()

                    continue

                elif status == "not_converging":
                    print("Status:", status, "  Not submitting.")
                    sys.stdout.flush()
                    continue

                elif status != "incomplete":
                    raise VaspWrapperError("unexpected convergence status: '" +
                                           status + "' and task: '" + task +
                                           "'")
                    # sys.stdout.flush()        ### Liz: Why is this here? Its unreachable.
                    # continue

                print("Preparing to submit a VASP Convergence job")
                sys.stdout.flush()

                # cd to configdir, submit jobs from configdir, then cd back to currdir
                currdir = os.getcwd()
                os.chdir(propdir)

                # determine the number of atoms in the configuration
                print("  Counting atoms in the POSCAR")
                sys.stdout.flush()
                pos = vasp.io.Poscar(os.path.join(self.configdir, "POS"))
                N = len(pos.basis)  #pylint: disable=invalid-name

                # Construct the run command
                cmd = ""
                if self.settings["preamble"] is not None:
                    # Append any instructions given in the 'preamble' file, if given
                    preamble = self.casm_directories.settings_path_crawl(
                        self.settings["preamble"], self.configname, self.clex)
                    with open(preamble, 'rb') as my_preamble:
                        cmd += "".join(my_preamble.read().decode('utf-8'))
                # Or just execute a single prerun line, if given
                if self.settings["prerun"] is not None:
                    cmd += self.settings["prerun"] + "\n"
                cmd += "python -c \"import casm.vaspwrapper; casm.vaspwrapper.Converge(configdir='" + self.configdir + "', propdir='" + propdir + "', prop=" + str(
                    prop) + ").run()\"\n"

                if self.settings["postrun"] is not None:
                    cmd += self.settings["postrun"] + "\n"

                print("  Constructing a job")
                sys.stdout.flush()

                # construct a Job
                job = Job(name=wrapper.jobname(self.configname) + "_" + '.'.join(propdir.split(os.sep)[-2:]),\
                              account=self.settings["account"],\
                              nodes=int(ceil(float(N)/float(self.settings["atom_per_proc"])/float(self.settings["ppn"]))),\
                              ppn=int(self.settings["ppn"]),\
                              walltime=self.settings["walltime"],\
                              pmem=self.settings["pmem"],\
                              qos=self.settings["qos"],\
                              queue=self.settings["queue"],\
                              message=self.settings["message"],\
                              email=self.settings["email"],\
                              priority=self.settings["priority"],\
                              constraint=self.settings["constraint"],\
                              command=cmd,\
                              auto=self.auto)

                print("  Submitting")
                sys.stdout.flush()
                # submit the job
                job.submit()
                self.report_status("submitted")

                # return to current directory and restore settings
                os.chdir(currdir)
                self.propdir = store_propdir
                self.prop = store_prop

            except UserWarning:
                continue

        print("CASM VASPWrapper Convergence job submission complete\n")
        sys.stdout.flush()

    def run_settings(self):
        """ Set default values based on runtime environment"""
        settings = dict(self.settings)

        # set default values

        if settings["npar"] == "CASM_DEFAULT":
            if "PBS_NUM_NODES" in os.environ:
                settings["npar"] = int(os.environ["PBS_NUM_NODES"])
            elif "SLURM_JOB_NUM_NODES" in os.environ:
                settings["npar"] = int(os.environ["SLURM_JOB_NUM_NODES"])
            else:
                settings["npar"] = None
        elif settings["npar"] == "VASP_DEFAULT":
            settings["npar"] = None

        if settings["npar"] is None:
            if settings["ncore"] == "CASM_DEFAULT":
                if "PBS_NUM_PPN" in os.environ:
                    settings["ncore"] = int(os.environ["PBS_NUM_PPN"])
                elif "SLURM_CPUS_ON_NODE" in os.environ:
                    settings["ncore"] = int(os.environ["SLURM_CPUS_ON_NODE"])
                else:
                    settings["ncore"] = None
            elif settings["ncore"] == "VASP_DEFAULT":
                settings["ncore"] = 1
        else:
            settings["ncore"] = None

        if settings["ncpus"] is None or settings["ncpus"] == "CASM_DEFAULT":
            if "PBS_NP" in os.environ:
                settings["ncpus"] = int(os.environ["PBS_NP"])
            elif "SLURM_NTASKS" in os.environ:
                settings["ncpus"] = int(os.environ["SLURM_NTASKS"])
            else:
                settings["ncpus"] = None

        if settings["run_limit"] is None or settings[
                "run_limit"] == "CASM_DEFAULT":
            settings["run_limit"] = 10

        return settings

    def run(self):
        """ Setup input files, run a vasp convergence, and report results """

        # construct the Converge object
        convergence = vasp.Converge(self.propdir, self.run_settings(),
                                    self.prop)

        # check the current status
        (status, task) = convergence.status()

        if status == "complete":
            print("Status:", status)
            sys.stdout.flush()

            # mark job as complete in db
            if self.auto:
                try:
                    complete_job()
                except (JobsError, JobDBError, EligibilityError) as my_error:
                    print(str(my_error))
                    sys.stdout.flush()

            # write results to properties.calc.json
            self.finalize()
            return

        elif status == "not_converging":
            print("Status:", status)
            self.report_status("failed", "run_limit")
            print("Returning")
            sys.stdout.flush()
            return

        elif status == "incomplete":

            if task == "setup":
                self.setup()

            self.report_status("started")
            (status, task) = convergence.run()

        else:
            self.report_status("failed", "unknown")
            raise VaspWrapperError("unexpected convergence status: '" +
                                   status + "' and task: '" + task + "'")
            # sys.stdout.flush()

        # once the run is done, update database records accordingly

        if status == "not_converging":

            # mark error
            if self.auto:
                try:
                    error_job("Not converging")
                except (JobsError, JobDBError) as my_error:
                    print(str(my_error))
                    sys.stdout.flush()

            print("Not Converging!")
            sys.stdout.flush()
            self.report_status("failed", "run_limit")

            # print a local settings file, so that the run_limit can be extended if the
            #   convergence problems are fixed
            try:
                os.makedirs(
                    os.path.join(self.configdir, "settings",
                                 self.casm_settings["curr_calctype"],
                                 self.settings["prop"] + "_converge"))
            except OSError:
                pass
            settingsfile = os.path.join(self.configdir, "settings",
                                        self.casm_settings["curr_calctype"],
                                        self.settings["prop"] + "_converge",
                                        "converge.json")
            write_settings(self.settings, settingsfile)

            print("Writing:", settingsfile)
            print("Edit the 'run_limit' property if you wish to continue.")
            sys.stdout.flush()
            return

        elif status == "complete":

            # mark job as complete in db
            if self.auto:
                try:
                    complete_job()
                except (JobsError, JobDBError, EligibilityError) as my_error:
                    print(str(my_error))
                    sys.stdout.flush()

            # write results to properties.calc.json
            self.finalize()

        else:
            self.report_status("failed", "unknown")
            raise VaspWrapperError(
                "vasp convergence complete with unexpected status: '" +
                status + "' and task: '" + task + "'")
            #sys.stdout.flush()

    def report_status(self, status, failure_type=None):
        """Report calculation status to status.json file in configuration directory.

        Args:
            status: string describing calculation status. Currently used values are
                 not_submitted
                 submitted
                 complete
                 failed
             failure_type: optional string describing reason for failure. Currently used values are
                 unknown
                 electronic_convergence
                 run_limit"""

        output = dict()
        output["status"] = status
        if failure_type is not None:
            output["failure_type"] = failure_type

        outputfile = os.path.join(self.propdir, "status.json")
        with open(outputfile, 'wb') as my_file:
            my_file.write(
                six.u(
                    json.dumps(output,
                               cls=noindent.NoIndentEncoder,
                               indent=4,
                               sort_keys=True)).encode('utf-8'))
        print("Wrote " + outputfile)
        sys.stdout.flush()

    def finalize(self):
        """ Checks if a series of runs is converged, and writes properties.calc.json as appropriate """
        if self.is_converged():
            # write properties.calc.json
            vaspdir = os.path.join(self.propdir, "run.final")
            structurefile = os.path.join(self.configdir, "POS")
            speciesfile = self.casm_directories.settings_path_crawl(
                "SPECIES", self.configname, self.clex)
            output = self.properties(vaspdir, structurefile, speciesfile)
            prop_file_name = "properties.calc.json"
            outputfile = os.path.join(self.propdir, prop_file_name)
            with open(outputfile, 'wb') as my_file:
                my_file.write(
                    six.u(
                        json.dumps(output,
                                   cls=noindent.NoIndentEncoder,
                                   indent=4,
                                   sort_keys=True)).encode('utf-8'))
            print("Wrote " + outputfile)
            sys.stdout.flush()
            self.report_status('complete')

    def is_converged(self):
        """Check for electronic convergence in completed calculations. Returns True or False."""

        # Verify that the last relaxation reached electronic convergence
        convergence = vasp.Converge(self.propdir, self.run_settings())
        for i in range(len(convergence.rundir)):
            try:
                print(self.propdir)
                vrun = vasp.io.Vasprun(
                    os.path.join(self.propdir, convergence.rundir[-i - 1],
                                 "vasprun.xml"))
                if len(vrun.all_e_0[-1]) >= vrun.nelm:
                    print((
                        'The last convergence run (' +
                        os.path.basename(convergence.rundir[-i - 1]) +
                        ') failed to achieve electronic convergence; properties.calc.json will not be written.\n'
                    ))
                    self.report_status('failed', 'electronic_convergence')
                    return False
                break
            ### WHY IS THIS HERE
            except:  #pylint: disable=bare-except
                pass

        # Verify that the final static run reached electronic convergence
        vrun = vasp.io.Vasprun(
            os.path.join(self.propdir, "run.final", "vasprun.xml"))
        if len(vrun.all_e_0[0]) >= vrun.nelm:
            print(
                'The final run failed to achieve electronic convergence; properties.calc.json will not be written.\n'
            )
            self.report_status('failed', 'electronic_convergence')
            return False

        return True

    def _settings_checker(self):
        """ Checks that settings given in converge.json are valid """
        # Check for required settings having sane values
        try:
            if not self.settings["prop"].upper() in VALID_PROP_TYPES:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): \"prop: %s\" not a valid convergence prop type!\nCurrently supported convergence prop types are %s"
                    % (self.settings["prop"], VALID_PROP_TYPES))
        except:
            raise ConvergeError(
                "Error in casm.vaspwrapper.Converge(): \"prop: %s\" missing, or could not be converted to a string!\nCurrently supported convergence prop types are %s"
                % (self.settings["prop"], VALID_PROP_TYPES))
        # ENCUT requires an int for start, stop, and step
        if self.settings["prop"].upper() == "ENCUT":
            try:
                self.settings["prop_start"] = int(self.settings["prop_start"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain integer prop_start for prop ENCUT. I found: %s"
                    % self.settings["prop_start"])
            try:
                self.settings["prop_step"] = int(self.settings["prop_step"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain integer prop_step for prop ENCUT. I found: %s"
                    % self.settings["prop_step"])
            try:
                self.settings["prop_stop"] = int(self.settings["prop_stop"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain integer prop_stop for prop ENCUT. I found: %s"
                    % self.settings["prop_step"])
        # NBANDS requires an int for start, stop, and step
        elif self.settings["prop"].upper() == "NBANDS":
            try:
                self.settings["prop_start"] = int(self.settings["prop_start"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain integer prop_start for prop NBANDS. I found: %s"
                    % self.settings["prop_start"])
            try:
                self.settings["prop_step"] = int(self.settings["prop_step"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain integer prop_step for prop NBANDS. I found: %s"
                    % self.settings["prop_step"])
            try:
                self.settings["prop_stop"] = int(self.settings["prop_stop"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain integer prop_stop for prop NBANDS. I found: %s"
                    % self.settings["prop_step"])
        # SIGMA requires a float for start, stop, and step
        elif self.settings["prop"].upper() == "SIGMA":
            try:
                self.settings["prop_start"] = float(
                    self.settings["prop_start"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain float prop_start for prop SIGMA. I found: %s"
                    % self.settings["prop_start"])
            try:
                self.settings["prop_step"] = float(self.settings["prop_step"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain float prop_step for prop SIGMA. I found: %s"
                    % self.settings["prop_step"])
            try:
                self.settings["prop_stop"] = float(self.settings["prop_stop"])
            except:
                raise ConvergeError(
                    "Error in casm.vaspwrapper.Converge(): converge.json must contain float prop_stop for prop SIGMA. I found: %s"
                    % self.settings["prop_step"])
        # KPOINTS requires either an int or a length-3 list of ints for start, stop, and step
        elif self.settings["prop"].upper() == "KPOINTS":
            try:
                self.settings["prop_start"] = int(self.settings["prop_start"])
            except TypeError:
                try:
                    self.settings["prop_start"] = [
                        int(k) for k in self.settings["prop_start"]
                    ]
                except:
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): converge.json must contain integer or 3-list of integer prop_start for prop KPOINTS. I found: %s"
                        % self.settings["prop_start"])
            if isinstance(self.settings["prop_start"], list):
                try:
                    self.settings["prop_step"] = [
                        int(k) for k in self.settings["prop_step"]
                    ]
                except:
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): converge.json must contain prop_step of the same type as prop_start for prop KPOINTS. I found: %s"
                        % self.settings["prop_step"])
                try:
                    self.settings["prop_stop"] = [
                        int(k) for k in self.settings["prop_stop"]
                    ]
                except:
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): converge.json must contain prop_stop of the same type as prop_start for prop KPOINTS. I found: %s"
                        % self.settings["prop_stop"])
            else:
                try:
                    self.settings["prop_step"] = int(
                        self.settings["prop_step"])
                except:
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): converge.json must contain prop_step of the same type as prop_start for prop KPOINTS. I found: %s"
                        % self.settings["prop_step"])
                try:
                    self.settings["prop_stop"] = int(
                        self.settings["prop_stop"])
                except:
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): converge.json must contain prop_stop of the same type as prop_start for prop KPOINTS. I found: %s"
                        % self.settings["prop_stop"])

        # If "tol" is present, check for a valid tol type and "tol_amount" value
        if self.settings["tol"] is not None:
            if isinstance(self.settings["tol"], list):
                if len(self.settings["tol"]) != len(
                        self.settings["tol_amount"]) or not isinstance(
                            self.settings["tol_amount"], list):
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): \"tol\" and \"tol_amount\" must have the same number of entries! However, you have %s and %s"
                        % (str(self.settings["tol"]),
                           str(self.settings["tol_amount"])))
                for my_tol, my_tol_amount in zip(self.settings["tol"],
                                                 self.settings["tol_amount"]):
                    try:
                        if not my_tol.lower() in VALID_TOL_TYPES:
                            raise ConvergeError(
                                "Error in casm.vaspwrapper.Converge(): \"tol: %s\" not a valid convergence tolerance type!\nCurrently supported convergence tolerance types are %s"
                                % (self.settings["tol"], VALID_TOL_TYPES))
                    except:
                        raise ConvergeError(
                            "Error in casm.vaspwrapper.Converge(): \"tol: %s\" not a valid convergence tolerance type!\nCurrently supported convergence tolerance types are %s"
                            % (self.settings["tol"], VALID_TOL_TYPES))
                    try:
                        my_tol_amount = abs(float(my_tol_amount))
                    except:
                        raise ConvergeError(
                            "Error in casm.vaspwrapper.Converge(): \"tol_amount: %s\" cannot be converted to float, but a float is needed!"
                            % self.settings["tol_amount"])
            else:
                try:
                    if not self.settings["tol"].lower() in VALID_TOL_TYPES:
                        raise ConvergeError(
                            "Error in casm.vaspwrapper.Converge(): \"tol: %s\" not a valid convergence tolerance type!\nCurrently supported convergence tolerance types are %s"
                            % (self.settings["tol"], VALID_TOL_TYPES))
                    else:
                        self.settings["tol"] = [self.settings["tol"]]
                except:
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): \"tol: %s\" not a valid convergence tolerance type!\nCurrently supported convergence tolerance types are %s"
                        % (self.settings["tol"], VALID_TOL_TYPES))

                try:
                    self.settings["tol_amount"] = [
                        abs(float(self.settings["tol_amount"]))
                    ]
                except:
                    raise ConvergeError(
                        "Error in casm.vaspwrapper.Converge(): \"tol_amount: %s\" cannot be converted to float, but a float is needed!"
                        % self.settings["tol_amount"])

    @property
    def configdir(self):
        """ Produces a configname from the configdir """
        return self.casm_directories.configuration_dir(self.configname)

    @staticmethod
    def properties(vaspdir, initial_structurefile=None, speciesfile=None):
        """Report results to properties.calc.json file in configuration directory, after checking for electronic convergence."""

        output = dict()
        vrun = vasp.io.Vasprun(os.path.join(vaspdir, "vasprun.xml"))

        # the calculation is run on the 'sorted' POSCAR, need to report results 'unsorted'

        if (initial_structurefile is not None) and (speciesfile is not None):
            species_settings = vasp.io.species_settings(speciesfile)
            initial_structure = vasp.io.Poscar(initial_structurefile,
                                               species_settings)
            unsort_dict = initial_structure.unsort_dict()
        else:
            # fake unsort_dict (unsort_dict[i] == i)
            unsort_dict = dict(
                zip(range(0, len(vrun.basis)), range(0, len(vrun.basis))))
        structure = vasp.io.Poscar(os.path.join(vaspdir, "POSCAR"))

        # unsort_dict:
        #   Returns 'unsort_dict', for which: unsorted_dict[orig_index] == sorted_index;
        #   unsorted_dict[sorted_index] == orig_index
        #   For example:
        #     'unsort_dict[0]' returns the index into the unsorted POSCAR of the first atom in the sorted POSCAR

        output["atom_type"] = structure.type_atoms
        output["atoms_per_type"] = structure.num_atoms
        output["coord_mode"] = vrun.coord_mode

        # as lists
        output["relaxed_forces"] = [None for i in range(len(vrun.forces))]
        for i, v in enumerate(vrun.forces):  #pylint: disable=invalid-name
            output["relaxed_forces"][unsort_dict[i]] = noindent.NoIndent(
                vrun.forces[i])

        output["relaxed_lattice"] = [
            noindent.NoIndent(v) for v in vrun.lattice
        ]

        output["relaxed_basis"] = [None for i in range(len(vrun.basis))]
        for i, v in enumerate(vrun.basis):  #pylint: disable=invalid-name
            output["relaxed_basis"][unsort_dict[i]] = noindent.NoIndent(
                vrun.basis[i])

        output["relaxed_energy"] = vrun.total_energy

        return output

    @staticmethod
    def volume(lat):
        """ Computes the volume of a parallelpiped given a 3-list of 3-lists (i.e., 3 3-vectors)"""
        try:
            return (lat[0][0] *
                    (lat[1][1] * lat[2][2] - lat[1][2] * lat[2][1]) -
                    lat[0][1] *
                    (lat[1][0] * lat[2][2] - lat[1][2] * lat[2][0]) +
                    lat[0][2] *
                    (lat[1][0] * lat[2][1] - lat[1][1] * lat[2][0]))
        except:
            raise RuntimeError(
                "The given lattice %s could not be parsed as a 3-list of 3-vectors."
                % str(lat))

    @staticmethod
    def lengths(lat):
        """ Computes the lengths of the 3 basis vectors in a lattice """
        try:
            return [
                sqrt(lat[0][0]**2 + lat[0][1]**2 + lat[0][2]**2),
                sqrt(lat[1][0]**2 + lat[1][1]**2 + lat[1][2]**2),
                sqrt(lat[2][0]**2 + lat[2][1]**2 + lat[2][2]**2)
            ]
        except:
            raise RuntimeError(
                "The given lattice %s could not be parsed as a 3-list of 3-vectors."
                % str(lat))

    @staticmethod
    def frange(x, y, jump=1.0):  #pylint: disable=invalid-name
        """Robust Range for floats, without using numpy."""
        i = 0.0
        x = float(x)  # Prevent yielding integers.
        x0 = x  #pylint: disable=invalid-name
        epsilon = jump / 2.0
        yield x  # yield always first value
        while x + epsilon < y:
            i += 1.0
            x = x0 + i * jump
            yield x



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import os, re, gzip


class OszicarError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class Oszicar(object):
    """Parse OSZICAR files.

       Currently, just contains:
           self.E = list of ionic step E0 values
           self.mag = list of ionic step magnetic moment values
           self.nelm = list of number of electronic steps in each ionic step
    """
    def __init__(self, filename):
        self.filename = filename
        self.E = []
        self.mag = []
        self.num_elm = []
        try:
            self.read()
        except OszicarError as e:
            raise e

    def read(self):
        """Parse OSZICAR file. Currently just collects E0 as list 'self.E'"""
        if os.path.isfile(self.filename):
            if self.filename.split(".")[-1].lower() == "gz":
                f = gzip.open(self.filename)
            else:
                f = open(self.filename)
        elif os.path.isfile(self.filename + ".gz"):
            f = gzip.open(self.filename + ".gz")
            raise OszicarError("file not found: " + self.filename)

        self.E = []
        for line in f:
            if re.search("E0", line):
                self.E.append(float(line.split()[4]))
                self.num_elm.append(int(prev_line.split()[1]))
            if re.search("mag", line):
                self.mag.append(float(line.split()[-1]))
            prev_line = str(line)  ## previous line
        f.close()



"""An interface to CASM projects via Python"""
from casm.project.project import project_path, ClexDescription, ProjectSettings, \
    DirectoryStructure, Project, Prim, CompositionAxes
from casm.project.selection import Selection
from casm.project.query import query
from casm.project.io import write_eci
from casm.project.attribute_info import AttributeInfo
__all__ = [
    'project_path', 'ClexDescription', 'ProjectSettings', 'DirectoryStructure',
    'Project', 'Prim', 'Selection', 'query', 'write_eci', 'AttributeInfo'
]



import numpy as np
from casm.misc import matrix


class OrbitalOccupationError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message


class OrbitalOccupation(object):
    """Orbital occupation information for a single site. Contains:
        spin_polarized : bool, whether the orbital occupation is spin-polarized
        l_quantum_number : int, the l quantum number of the specified orbitals
        matrices : list, one or two numpy arrays corresponding to the occupation matrices of each spin channel
    """
    def __init__(self, occupation_matrix_a, occupation_matrix_b=None):
        """Initialize occupation from one or two matrices,
            which can be given as lists of lists or numpy arrays
        """
        self.spin_polarized = False
        occupation_matrix_a = np.array(occupation_matrix_a)
        ma, na = occupation_matrix_a.shape
        if (ma == na and na % 2 == 1):
            self.l_quantum_number = int((na - 1) / 2)
            self.matrices = [occupation_matrix_a]
        else:
            raise OrbitalOccupationError(
                'Occupation matrix a is the wrong shape')
        if occupation_matrix_b is not None:
            occupation_matrix_b = np.array(occupation_matrix_b)
            self.spin_polarized = True
            mb, nb = occupation_matrix_b.shape
            if (mb == nb and nb == na):
                self.matrices.append(occupation_matrix_b)
            else:
                raise OrbitalOccupationError(
                    'Occupation matrices a and b have different shapes')

    def get_occext_string(self, site_number):
        """Output a string containing the orbital occupation information using the OCCMATRIX format specified by
            https://github.com/WatsonGroupTCD/Occupation-matrix-control-in-VASP/blob/master/README.md
            site_number : int, POSCAR site number to be written
        """
        def matrix_to_string(matrix):
            return '\n'.join('\t'.join(f'{entry:.4f}' for entry in row)
                             for row in matrix)

        output_string = f'{site_number}\t{self.l_quantum_number}\t{int(self.spin_polarized)+1}\n'
        output_string += 'spin channel 1\n'
        output_string += matrix_to_string(self.matrices[0]) + '\n'
        if self.spin_polarized:
            output_string += 'spin channel 2\n'
            output_string += matrix_to_string(self.matrices[1]) + '\n'
        output_string += '\n'
        return output_string

    def to_vector(self):
        """Return a vector of the unrolled occupation matrix/matrices that can be read as a CASM attribute.
            In the spin-polarized case, the vectors corresponding to the up/down occupation matrices are concatenated.
        """
        vector = []
        for occupation_matrix in self.matrices:
            vector += matrix.unroll_symmetric_matrix(occupation_matrix)
        return vector


def write_occupations(filename, occupations):
    """Write a dict of occupations out to an OCCMATRIX-like file
        filename : str, output filename
        occupations : dict of OrbitalOccupation, site occupations stored using site indices (from 0) as keys
    """
    with open(filename, 'w') as output_file:
        output_file.write(str(len(occupations.keys())) + '\n')
        for site_number in occupations.keys():
            output_file.write(occupations[site_number].get_occext_string(
                site_number + 1))  # VASP indexes sites from 1



# # conda's current version of pandas raises these warnings, but they are safe
# # see: https://stackoverflow.com/questions/40845304
# import warnings
# warnings.filterwarnings("ignore", message="numpy.dtype size changed")
# warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

from io import StringIO
import pandas
import casm
from casm.misc import compat


def query(proj,
          columns,
          selection_path,
          selection_type,
          verbatim=True,
          all=False):
    """Return a pandas.DataFrame with `casm query` output

    Arguments
    ---------
    proj: casm.project.Project
        The project to query

    columns: iterable of str
        Corresponds to `casm query -k` arguments listing the values to be queried. Use `casm query --help properties` and `casm query --help operators` to list options.

    selection_path: str
        The `-c,--selection` option, a path to a selection file, or a standard selection name ("MASTER", "ALL", "CALCULATED", "NONE")

    selection_type: str
        The `-t,--type` option, indicates the type of object being selected. Expected to be one of:

            "config": to select configurations
            "scel": to select supercells

    verbatim: bool
        If True, use `-v,--verbatim` option to exclude 'name' and 'selected' columns from the query output.

    all: bool
        If True, use `-a,--all` to include unselected objects in the output

    Returns
    -------
    data: pandas.DataFrame
        A DataFrame containing the query results. Note that no columns are loaded as bool dtype.
    """
    args = query_args(columns, selection_path, selection_type, verbatim, all)

    stdout, stderr, returncode = proj.capture(args)

    try:
        sout = StringIO(stdout)
        if compat.peek(sout) == '#':
            sout.read(1)
        return pandas.read_csv(sout, sep=compat.str(' +'), engine='python')
    except:
        print("Error in casm.query")
        print("  proj:", proj.path)
        print("  Attempted to execute: '" + args + "'")
        print("---- stdout: ---------------------")
        print(stdout)
        print("---- stderr: ---------------------")
        print(stderr)
        print("----------------------------------")
        raise


def query_args(columns,
               selection_path,
               selection_type,
               verbatim=True,
               all=False):
    """Constructs a string appropriate for use by `casm query`

    Arguments
    ---------
    columns: iterable of str
        Corresponds to `casm query -k` arguments listing the values to be queried. Use `casm query --help properties` and `casm query --help operators` to list options.

    selection_path: str
        The `-c,--selection` option, a path to a selection file, or a standard selection name ("MASTER", "ALL", "CALCULATED", "NONE")

    selection_type: str
        The `-t,--type` option, indicates the type of object being selected. Expected to be one of:

            "config": to select configurations
            "scel": to select supercells

    verbatim: bool
        If True, use `-v,--verbatim` option to exclude 'name' and 'selected' columns from the query output.

    all: bool
        If True, use `-a,--all` to include unselected objects in the output

    Returns
    -------
    args: str
        A string appropriate for use by `casm query`. The `-o STDOUT` option is always included. Example:

            "query -k 'comp_n scel_size' -c MASTER -t config -v -o STDOUT"
    """
    args = "query -k "
    args += "'"
    for k in columns:
        args += k + " "
    args += "'"
    args += " -c " + selection_path
    args += " -t " + selection_type
    if verbatim == True:
        args += " -v"
    if all:
        args += " -a"
    args += " -o STDOUT"
    return args



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import numpy as np
import math
import os.path
import json


class PoscarError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class Site:
    """ Site in a basis.

        Contains:
            self.cart = True if cartesian coordinate
            self.SD_FLAG = string for selective dynamics description, empty string by default (ex. "T F F").
            self.occupant = CASM specie name, empty string by default
            self.occ_alias = alias (POTCAR) name, empty string by default
            self.position = np.array coordinate
    """
    def __init__(self, cart, position, SD_FLAG="", occupant="", occ_alias=""):
        """ Site constructor """
        self.cart = cart
        self.SD_FLAG = SD_FLAG
        self.occupant = occupant
        self.occ_alias = occ_alias
        if not isinstance(position, np.ndarray):
            raise PoscarError(
                "Attempting to construct a Site and 'position' is not a numpy ndarray"
            )
        self.position = position

    def write(self, file):
        """ Write this Site to a file """
        np.savetxt(file, self.position, fmt='%.8f', newline='    ')
        file.write(self.SD_FLAG)
        file.write(self.occupant)
        file.write('\n')


class Poscar:
    """
    The Poscar class contains:
        self.header: Contains POSCAR header line with whitespace stripped from beginning and end
        self.scaling: lattice scale
        self._lattice: the lattice; lattice vectors are stored as rows in a numpy array
        self._reciprocal_lattice: the reciprocal lattice; lattice vectors are stored as rows in a numpy array
        self.SD_FLAG: True or False, Selective Dynamics Flag
        self.type_atoms: lists the specie names as in the POS (ex. [Mn3 Mn4])
        self.type_atoms_alias: lists the POTCAR names for the species
        self.num_atoms: lists the atoms as in the POS (ex. [1 1])
        self.coord_mode: Contains the coordinate mode text from POSCAR, with whitespace stripped from beginning and end
        self.basis: a list of Site objects

    """
    def __init__(self, filename, species=None, legacy_support=True):
        """ Construct a Poscar object from 'filename'

            Args:
                filename = POS/POSCAR file or structure.json file to read
                species = (default None) If given a Species dict, it is used to set self.type_atoms_alias for determining which POTCARs to use
        """
        if os.path.splitext(filename)[1] == '.json':
            self.read_structure_json(filename, species)
        else:
            self.read(filename, species, legacy_support)

    def read(self, filename, species=None, legacy_support=True):
        """ Reads a POS/POSCAR from 'filename'"""
        self.basis = []
        self._lattice = np.zeros((3, 3))
        self._reciprocal_lattice = np.zeros((3, 3))
        self.coord_mode = ''
        self.SD_FLAG = False
        self.num_atoms = []
        self.type_atoms = []
        self.type_atoms_alias = list(self.type_atoms)
        self.header = ""
        self.scaling = 0
        try:
            file = open(filename, 'r')
        except IOError:
            raise PoscarError("Could not read file: " + filename)
        self.header = file.readline().strip()
        try:
            self._read_lattice(file)
            self._read_atominfo(file)
            self._read_flags(file)
            self._read_basis(file)
        except PoscarError as e:
            if legacy_support == False:
                raise e
            try:
                file.seek(0)
                file.readline()
                self._read_lattice(file)
                self._read_atominfo_legacy(file)
                self._read_flags(file)
                self._read_basis_legacy(file)
            except PoscarError as e:
                raise e
        file.close()

        # set type alias
        if species != None:
            self.update(species)

    def read_structure_json(self, filename, species=None):
        """
            Reads a structure from structure.json file 'filename'
        """
        self.header = ""
        try:
            file = open(filename, 'r')
        except IOError:
            raise PoscarError("Could not read file: " + filename)
        config_data = json.loads(file.read())
        file.close()

        # read lattice
        self.scaling = 1.0
        self._lattice = np.array(config_data['lattice'])
        self._compute_reciprocal_lattice()

        # read basis
        self.SD_FLAG = False
        self.coord_mode = config_data['coord_mode']
        self.basis = []
        cart = self.coord_mode[0].lower() == 'c'
        self.num_atoms = []
        self.type_atoms = []

        # sort atom_coords and atom_types based on atom_types
        zip_list = list(
            zip(config_data['atom_coords'], config_data['atom_type']))
        zip_list.sort(key=lambda x: x[1])
        sorted_atom_types = [x[1] for x in zip_list]
        sorted_atom_coords = [x[0] for x in zip_list]

        for i, pos in enumerate(sorted_atom_coords):
            atom_name = sorted_atom_types[i]
            if not atom_name in self.type_atoms:
                self.type_atoms.append(atom_name)
                self.num_atoms.append(0)
            self.num_atoms[-1] += 1
            self.basis.append(
                Site(cart, np.array(pos), '', atom_name, atom_name))
        self.type_atoms_alias = list(self.type_atoms)

        # set type alias
        if species != None:
            self.update(species)

        # Storing the atom_type as listed in structure.json which can be used while printing out
        self.atom_type = sorted_atom_types

    def write(self, filename, sort=True):
        """ Write Poscar to 'filename'.

            If 'sort' = True, sort basis sites to minimize number of POTCARs
        """
        try:
            file = open(filename, 'w')
        except IOError:
            raise PoscarError("Could not write: " + filename)
        file.write(str(self.header) + '\n')
        file.write(str(self.scaling) + '\n')
        np.savetxt(file, self._lattice, newline='\n', fmt='%.8f')

        if sort:
            type_line = ''
            num_line = ''
            d = self.basis_dict()
            for atom in sorted(d.keys()):
                type_line = type_line + '   ' + atom
                num_line = num_line + '   ' + str(len(d[atom]))
            if (self.SD_FLAG):
                file.write('Selective Dynamics\n')
            file.write(type_line + '\n')
            file.write(num_line + '\n')
            file.write(self.coord_mode + '\n')
            for atom in sorted(d.keys()):
                for tb in d[atom]:
                    tb.write(file)
        else:
            file.write(' '.join(self.type_atoms_alias) + '\n')
            file.write(' '.join([str(x) for x in self.num_atoms]) + '\n')
            if (self.SD_FLAG):
                file.write('Selective Dynamics\n')
            file.write(self.coord_mode + '\n')
            for s in self.basis:
                s.write(file)

        file.close()

    def lattice(self, index=None):
        """ Returns the lattice, or lattice vector 'index', as numpy array"""
        if index != None:
            return self._lattice[index, :]
        return self._lattice

    def reciprocal_lattice(self, index=None):
        """ Returns the reciprocal lattice vector 'index', as numpy array"""
        if index != None:
            return self._reciprocal_lattice[index, :]
        return self._reciprocal_lattice

    def volume(self, lattice=None):
        """ Returns scalar triple product of lattice vectors """
        if lattice is None:
            lattice = self._lattice
        return np.inner(lattice[0, :], np.cross(lattice[1, :], lattice[2, :]))

    def reciprocal_volume(self, reciprocal_lattice=None):
        """ Returns scalar triple product of reciprocal lattice vector """
        if reciprocal_lattice is None:
            reciprocal_lattice = self._reciprocal_lattice
        return self.volume(reciprocal_lattice)

    def basis_dict(self):
        """ Return a dictionary where keys are unique specie 'alias' and values are lists of atoms."""
        basis_dict = dict()
        for i, atom in enumerate(self.type_atoms_alias):
            start = sum(self.num_atoms[0:(i)])
            end = start + self.num_atoms[i]
            if atom not in basis_dict.keys():
                basis_dict[atom] = self.basis[start:end]
            else:
                basis_dict[atom] += self.basis[start:end]
        return basis_dict

    def unsort_dict(self):
        """ Return a dict to unsort atom order.

            Returns 'unsort_dict', for which: unsorted_dict[orig_index] == sorted_index;

            unsorted_dict[sorted_index] == orig_index

            For example:
              'unsort_dict[0]' returns the index into the unsorted POSCAR of the first atom in the sorted POSCAR
        """

        # create basis_dict, but with initial position in POSCAR instead of coordinate
        basis_dict = dict()
        for i, atom in enumerate(self.type_atoms_alias):
            start = sum(self.num_atoms[0:(i)])
            end = start + self.num_atoms[i]
            if atom not in basis_dict.keys():
                basis_dict[atom] = range(start, end)
            else:
                basis_dict[atom] += range(start, end)
        # print(basis_dict)

        orig_pos = []
        for atom in sorted(basis_dict.keys()):
            orig_pos += basis_dict[atom]

        # print(orig_pos)

        new_pos = range(0, len(self.basis))

        return dict(zip(new_pos, orig_pos))

    def _compute_reciprocal_lattice(self):
        """ Compute the reciprocal lattice based on the current lattice and store it"""
        self._reciprocal_lattice = 2.0*math.pi * \
            np.linalg.inv(np.transpose(self._lattice))
        return

    def _read_lattice(self, file):
        """ Called by self.read() to read the lattice into self._lattice

            Args:
                file: an open POSCAR being read from

            self._lattice contains lattice vectors stored as rows in a numpy array (easy inversion)
        """
        try:
            line = file.readline().strip()
            self.scaling = float(line)
        except ValueError:
            raise PoscarError("Could not read lattice scaling: '" + line + "'")
        lat = []
        for i in range(3):
            line = file.readline()
            try:
                lat.append([float(x) for x in line.split()])
            except ValueError:
                raise PoscarError("Could not read lattice vector: '" + line +
                                  "'")
        self._lattice = self.scaling * np.array(lat)
        if self._lattice.shape != (3, 3):
            raise PoscarError("Lattice shape error: " +
                              np.array_str(self._lattice))
        self._compute_reciprocal_lattice()
        self.scaling = 1.0
        return

    def _read_atominfo(self, file):
        """ Called by self.read() to read the 2 lines that correspond to the num_atoms and type_atoms
            in a VASP5 style POSCAR

            Args:
                file: an open POSCAR being read from

            self.type_atoms is a list of strings corresponding to atom names
            self.num_atoms is a list of int corresponding to the number of each atom type
            self.type_atoms_alias is set from self.type_atoms
        """
        typeline = file.readline().strip()
        numline = file.readline().strip()
        self.type_atoms = typeline.split()
        if len(self.type_atoms) == 0:
            raise PoscarError("No atom names found: '" + typeline + "'")
        try:
            self.num_atoms = [int(n) for n in numline.split()]
        except ValueError:
            raise PoscarError("Could not read number of each atom type: '" +
                              numline + "'")
        if len(self.num_atoms) != len(self.type_atoms):
            raise PoscarError(
                "Atom type and number lists are not the same length: \n'" +
                typeline + "'\n" + "'" + numline + "'")
        self.type_atoms_alias = list(self.type_atoms)

    def _read_atominfo_legacy(self, file):
        """ Called by self.read() to read the lines that correspond to the num_atoms and type_atoms
            in a VASP4 style POSCAR

            Args:
                file: an open POSCAR being read from

            self.type_atoms is a list of strings corresponding to atom names
            self.num_atoms is a list of int corresponding to the number of each atom type
            self.type_atoms_alias is set from self.type_atoms
        """
        numline = file.readline().strip()
        try:
            self.num_atoms = [int(n) for n in numline.split()]
        except ValueError:
            raise PoscarError("Could not read number of each atom type: '" +
                              numline + "'")
        self.type_atoms = list(map(int, range(len(self.num_atoms))))
        self.type_atoms_alias = list(self.type_atoms)

    def _read_flags(self, file):
        """ Called by self.read() to read POSCAR flags: Selective Dynamics, and Coord Mode

            Args:
                file: an open POSCAR being read from

            self.SD_FLAG = True/False if selective dynamics on
            self.coord_mode = Contains the coordinate mode text, with whitespace stripped from beginning and end
        """
        line = file.readline().strip()
        if len(line) == 0:
            raise PoscarError(
                "Could not read Select Dynamics or Coord Mode line")
        if (line[0].lower() == 's'):
            self.SD_FLAG = True
            self.coord_mode = file.readline()
            if len(line) == 0:
                raise PoscarError(
                    "Read Select Dynamics, but could not read Coord Mode line")
        else:
            self.SD_FLAG = False
            self.coord_mode = line

        if self.coord_mode[0].lower() not in ['c', 'd', 'k']:
            raise PoscarError("Read invalid coord mode: '" + self.coord_mode +
                              "'")

    def _read_basis(self, file):
        """ Called by self.read() to read basis of POSCAR file into self.basis

            Args:
                file: an open POSCAR being read from

            self.basis is a list of Site objects
        """
        self.basis = []
        cart = not (self.coord_mode[0].lower() == 'd')
        for i in range(len(self.num_atoms)):
            for j in range(self.num_atoms[i]):
                line = file.readline().strip()
                if len(line) == 0:
                    raise PoscarError(
                        "Error reading basis: insufficient number of atoms")
                words = line.split()
                pos = []
                if (self.SD_FLAG):
                    if len(words) < 6:
                        raise PoscarError(
                            "Error reading basis: insufficient number of SD tags"
                        )
                    SD_FLAG = words[3] + ' ' + words[4] + ' ' + words[5]
                else:
                    SD_FLAG = ''
                try:
                    pos = [float(x) for x in words[0:3]]
                except ValueError:
                    raise PoscarError("Error reading basis coordinate: '" +
                                      line + "'")
                self.basis.append(
                    Site(cart, np.array(pos), SD_FLAG, self.type_atoms[i],
                         self.type_atoms[i]))

    def _read_basis_legacy(self, file):
        """ Called by self.read() to read basis of VASP4  POSCAR file into self.basis

            Args:
                file: an open POSCAR being read from

            self.basis is a list of Site objects
        """
        self.basis = []
        cart = not (self.coord_mode[0].lower() == 'd')
        for i in range(len(self.num_atoms)):
            for j in range(self.num_atoms[i]):
                line = file.readline().strip()
                if len(line) == 0:
                    raise PoscarError(
                        "Error reading basis: insufficient number of atoms")
                words = line.split()
                pos = []
                if (self.SD_FLAG):
                    if len(words) < 6:
                        raise PoscarError(
                            "Error reading basis: insufficient number of SD tags"
                        )
                    SD_FLAG = words[3] + ' ' + words[4] + ' ' + words[5]
                else:
                    SD_FLAG = ''
                try:
                    pos = [float(x) for x in words[0:3]]
                except ValueError:
                    raise PoscarError("Error reading basis coordinate: '" +
                                      line + "'")
                if len(words) == 7 or len(words) == 4:
                    atom_type = words[-1]
                    if j > 0:
                        if self.type_atoms[i] != atom_type:
                            raise PoscarError(
                                "Error reading basis: mismatch between type count in header and types in basis"
                            )
                    else:
                        self.type_atoms[i] = atom_type
                        self.type_atoms_alias[i] = atom_type
                else:
                    atom_type = self.type_atoms[i]
                self.basis.append(
                    Site(cart, np.array(pos), SD_FLAG, atom_type, atom_type))

    def update(self, species):
        """ Set self.type_atoms_alias and self.basis[x].alias according to Species dict.

            Warning: Not sure we want to check equivalence regardless of atom name case
        """
        for i, atom in enumerate(self.type_atoms):
            for s in species:
                if s == atom:
                    self.type_atoms_alias[i] = species[s].alias
        index = 0
        for i in range(len(self.type_atoms_alias)):
            for j in range(self.num_atoms[i]):
                self.basis[index].occ_alias = self.type_atoms_alias[i]
                index += 1

        if 'MAGMOM' in species.tags:
            for base in self.basis:
                if base.occupant in species.keys():
                    base.mag = float(species[base.occupant].tags['MAGMOM'])

        return

    def apply_deformation(self, deformation=np.ones((3, 3))):
        """ applies a deformation which is array of size 3 X 3
        """
        self._lattice = np.dot(self._lattice, np.array(deformation).T)



"""Defines the relax module methods"""
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import json
import math
import os
import re
import six
import sys
import json
import shutil
import casm
from casm.vaspwrapper.vasp_calculator_base import VaspCalculatorBase
from casm.vasp import Relax as calculator

from casm import vasp, wrapper
from casm.misc import noindent
from casm.project import DirectoryStructure, ProjectSettings
from casm.vaspwrapper import VaspWrapperError, read_settings, write_settings
from casm.vaspwrapper.vasp_calculator_base import VaspCalculatorBase


class Relax(VaspCalculatorBase):
    """The Relax class contains functions for setting up, executing, and parsing a VASP relaxation.

    Attributes
    ----------
    selection : casm.project.Selection
        selection of configuration
    calctype : string
        calctype to setup and run the neb calculations
    auto : bool
    sort : bool

    Methods
    -------
    from_configuration_dir(configuration_dir='string', calctype='string', bool, bool)
        returns a instance of the Neb class instantited with a single configuration
    config_properties(config_data=dict/Pandas.DataFrame)
        return a dict of the properties required to setup a configuration
    pre_setup
        creates folder and makes POS files for each image
    setup
        sets up the input vasp files for the selection
    config_setup
        sets up the input vasp files for a single configuration
    get_vasp_input_files(config_data=dict/Pandas.DataFrame, settings=dict)
        returns filenames of a vasp neb calculation
    submit
        submit a job for each configuration
    run
        runs the neb calcutation on the selection
    report
        reports results for the selection
    run_cmd(configdir='string', calctype='string')
        return a string of command to run a single configuration
    finalize(config_data=dict/pandas_data, super_poscarfile='string')
        checks convergnce and write a properties file for the selection
    properties(calcdir='string', super_poscarfile='string', speciesfile='string')
        return a dict containing all the relaxed properties for a configuration

    Notes
    -----
    The Relax class contains functions for setting up, executing, and parsing a VASP relaxation.

    The relaxation creates the following directory structure:
    config/
        calctype.name/
            run.0/
            ....

    'run.i' directories are only created when ready.

    This automatically looks for VASP settings files using:
        casm.project.DirectoryStructure.settings_path_crawl

    """
    def __init__(self, selection, calctype=None, auto=True, sort=True):
        """Construct a VASP neb job object"""
        print("Construct a casm.vaspwrapper.Relax instance:")
        VaspCalculatorBase.__init__(self, selection, calctype, auto, sort)
        self.calculator = calculator

    def pre_setup(self):
        """Setus up folders and writes POS files"""
        self.selection.write_pos()
        for index, config_data in self.selection.data.iterrows():
            try:
                os.makedirs(config_data["calcdir"])
            except:
                pass

    def config_setup(self, config_data):
        """ Setup initial relaxation run for a configuration

            Uses the following files from the most local .../settings/calctype.name directory:
                INCAR: VASP input settings
                KPOINTS: VASP kpoints settings
                POSCAR: reference for KPOINTS if KPOINTS mode is not A/AUTO/Automatic
                SPECIES: info for each species such as which POTCAR files to use, MAGMOM, GGA+U, etc.

            Uses the following files from the .../config_calcdir/00 directory:
                POSCAR: sample structure of the configuration to be relaxed

        """
        settings = self.read_settings(config_data["setfile"])
        super(Relax, self).config_setup(config_data)
        if settings["initial_deformation"] != None:
            deformation = self.get_deformation(settings)
            self.apply_deformation(deformation, config_data["calcdir"])

    def setup(self):
        """Setup initial relaxation run for the selection"""
        super(Relax, self).setup()

    def submit(self):
        """Submit a job for each configuration"""
        super(Relax, self).submit()

    def run(self):
        """Runs the neb calcutation on the selection"""
        super(Relax, self).run()

    def report(self):
        """Reports results for the selection"""
        super(Relax, self).report()

    @staticmethod
    def run_cmd(configdir, calctype):
        """Return a string of command to run a single configuration"""
        return "python -c \"import casm.vaspwrapper; obj = casm.vaspwrapper.Relax.from_configuration_dir('{0}', '{1}'); obj.run()\"\n".format(
            configdir, calctype)

    def finalize(self, config_data, super_poscarfile=None):
        """Checks convergnce and write a properties file for the selection"""
        if super_poscarfile is None:
            super_poscarfile = os.path.join(config_data["configdir"], "POS")
        super(Relax, self).finalize(config_data, super_poscarfile)
        sys.stdout.flush()

    def properties(self, calcdir, super_poscarfile=None, speciesfile=None):
        """Make properties output as a list of dict of each image properties"""
        output = super(Relax, self).properties(calcdir, super_poscarfile,
                                               speciesfile)
        return output



from __future__ import (absolute_import, division, print_function,
                        unicode_literals)
from builtins import *

import casm
import casm.vaspwrapper
import sys


def main():
    print("Begin vasp.setup")

    if len(sys.argv) != 2:
        print("Usage: vasp.setup configdir")
        sys.exit()

    configdir = sys.argv[1]

    relaxation = casm.vaspwrapper.Relax(configdir)
    relaxation.setup()

    print("Finish vasp.setup\n\n")


if __name__ == "__main__":
    main()
